#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional, TYPE_CHECKING, Union

from ..enums import SNAPSHOT_POLICY

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class SecondaryDatasetDict(TypedDict):
        identifier: str
        catalog_id: str
        catalog_version_id: str
        snapshot_policy: str


FeatureDerivationWindowsType = List[Dict[str, Union[int, str]]]
RelationshipDictType = Dict[
    str,
    Union[
        None,
        str,
        int,
        List[str],
        FeatureDerivationWindowsType,
    ],
]

__all__ = (
    "DatasetDefinition",
    "FeatureDiscoverySetting",
    "Relationship",
    "SecondaryDataset",
)


class DatasetDefinition:
    """Dataset definition for the Feature Discovery

    .. versionadded:: v2.25

    Attributes
    ----------
    identifier: string
        Alias of the dataset (used directly as part of the generated feature names)
    catalog_id: string, optional
        Identifier of the catalog item
    catalog_version_id: string
        Identifier of the catalog item version
    primary_temporal_key: string, optional
        Name of the column indicating time of record creation
    feature_list_id: string, optional
        Identifier of the feature list. This decides which columns in the dataset are
        used for feature generation
    snapshot_policy: string, optional
        Policy to use  when creating a project or making predictions.
        If omitted, by default endpoint will use 'latest'.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        dataset_definition = dr.DatasetDefinition(
            identifier='profile',
            catalog_id='5ec4aec1f072bc028e3471ae',
            catalog_version_id='5ec4aec2f072bc028e3471b1',
        )

        dataset_definition = dr.DatasetDefinition(
            identifier='transaction',
            catalog_id='5ec4aec1f072bc028e3471ae',
            catalog_version_id='5ec4aec2f072bc028e3471b1',
            primary_temporal_key='Date'
        )
    """

    def __init__(
        self,
        identifier: str,
        catalog_id: Optional[str],
        catalog_version_id: str,
        snapshot_policy: str = SNAPSHOT_POLICY.LATEST,
        feature_list_id: Optional[str] = None,
        primary_temporal_key: Optional[str] = None,
    ):
        self.identifier = identifier
        self.catalog_id = catalog_id
        self.catalog_version_id = catalog_version_id
        self.snapshot_policy = snapshot_policy
        self.primary_temporal_key = primary_temporal_key
        self.feature_list_id = feature_list_id

    def to_dict(self) -> Dict[str, Optional[str]]:
        return {
            "identifier": self.identifier,
            "catalog_id": self.catalog_id,
            "catalog_version_id": self.catalog_version_id,
            "snapshot_policy": self.snapshot_policy,
            "primary_temporal_key": self.primary_temporal_key,
            "feature_list_id": self.feature_list_id,
        }

    # pylint: disable-next=missing-function-docstring
    def to_payload(self) -> Dict[str, Optional[str]]:
        payload = {
            "identifier": self.identifier,
            "catalogId": self.catalog_id,
            "catalogVersionId": self.catalog_version_id,
            "snapshotPolicy": self.snapshot_policy,
        }
        if self.primary_temporal_key:
            payload["primaryTemporalKey"] = self.primary_temporal_key
        if self.feature_list_id:
            payload["featureListId"] = self.feature_list_id
        return payload


class Relationship:
    """Relationship between dataset defined in DatasetDefinition

    .. versionadded:: v2.25

    Attributes
    ----------
    dataset1_identifier: string, optional
        Identifier of the first dataset in this relationship.
        This is specified in the identifier field of dataset_definition structure.
        If None, then the relationship is with the primary dataset.
    dataset2_identifier: string
        Identifier of the second dataset in this relationship.
        This is specified in the identifier field of dataset_definition schema.
    dataset1_keys: list of string (max length: 10 min length: 1)
        Column(s) from the first dataset which are used to join to the second dataset
    dataset2_keys: list of string (max length: 10 min length: 1)
        Column(s) from the second dataset that are used to join to the first dataset
    feature_derivation_window_start: int, or None
        How many time_units of each dataset's primary temporal key into the past relative
        to the datetimePartitionColumn the feature derivation window should begin.
        Will be a negative integer,
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_end: int, optional
        How many timeUnits of each dataset's record
        primary temporal key into the past relative to the datetimePartitionColumn the
        feature derivation window should end.  Will be a non-positive integer, if present.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_time_unit: int, optional
        Time unit of the feature derivation window.
        One of ``datarobot.enums.AllowedTimeUnitsSAFER``
        If present, time-aware joins will be used.
        Only applicable when dataset1_identifier is not provided.
    feature_derivation_windows: list of dict, or None
        List of feature derivation windows settings. If present, time-aware joins will be used.
        Only allowed when feature_derivation_window_start,
        feature_derivation_window_end and feature_derivation_window_time_unit are not provided.
    prediction_point_rounding: int, optional
        Closest value of prediction_point_rounding_time_unit to round the prediction point
        into the past when applying the feature derivation window. Will be a positive integer,
        if present.Only applicable when dataset1_identifier is not provided.
    prediction_point_rounding_time_unit: string, optional
        Time unit of the prediction point rounding.
        One of ``datarobot.enums.AllowedTimeUnitsSAFER``
        Only applicable when dataset1_identifier is not provided.

    The `feature_derivation_windows` is a list of dictionary with schema:
        start: int
            How many time_units of each dataset's primary temporal key into the past relative
            to the datetimePartitionColumn the feature derivation window should begin.
        end: int
            How many timeUnits of each dataset's record
            primary temporal key into the past relative to the datetimePartitionColumn the
            feature derivation window should end.
        unit: string
            Time unit of the feature derivation window.
            One of ``datarobot.enums.AllowedTimeUnitsSAFER``.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        relationship = dr.Relationship(
            dataset1_identifier='profile',
            dataset2_identifier='transaction',
            dataset1_keys=['CustomerID'],
            dataset2_keys=['CustomerID']
        )

        relationship = dr.Relationship(
            dataset2_identifier='profile',
            dataset1_keys=['CustomerID'],
            dataset2_keys=['CustomerID'],
            feature_derivation_window_start=-14,
            feature_derivation_window_end=-1,
            feature_derivation_window_time_unit='DAY',
            prediction_point_rounding=1,
            prediction_point_rounding_time_unit='DAY'
        )
    """

    def __init__(
        self,
        dataset2_identifier: str,
        dataset1_keys: List[str],
        dataset2_keys: List[str],
        dataset1_identifier: Optional[str] = None,
        feature_derivation_window_start: Optional[int] = None,
        feature_derivation_window_end: Optional[int] = None,
        feature_derivation_window_time_unit: Optional[int] = None,
        feature_derivation_windows: Optional[FeatureDerivationWindowsType] = None,
        prediction_point_rounding: Optional[int] = None,
        prediction_point_rounding_time_unit: Optional[str] = None,
    ) -> None:
        self.dataset1_identifier = dataset1_identifier
        self.dataset2_identifier = dataset2_identifier
        self.dataset1_keys = dataset1_keys
        self.dataset2_keys = dataset2_keys
        self.feature_derivation_window_start = feature_derivation_window_start
        self.feature_derivation_window_end = feature_derivation_window_end
        self.feature_derivation_window_time_unit = feature_derivation_window_time_unit
        self.feature_derivation_windows = feature_derivation_windows
        self.prediction_point_rounding = prediction_point_rounding
        self.prediction_point_rounding_time_unit = prediction_point_rounding_time_unit

    def to_dict(self) -> RelationshipDictType:
        return {
            "dataset1_identifier": self.dataset1_identifier,
            "dataset2_identifier": self.dataset2_identifier,
            "dataset1_keys": self.dataset1_keys,
            "dataset2_keys": self.dataset2_keys,
            "feature_derivation_window_start": self.feature_derivation_window_start,
            "feature_derivation_window_end": self.feature_derivation_window_end,
            "feature_derivation_window_time_unit": self.feature_derivation_window_time_unit,
            "feature_derivation_windows": self.feature_derivation_windows,
            "prediction_point_rounding": self.prediction_point_rounding,
            "prediction_point_rounding_time_unit": self.prediction_point_rounding_time_unit,
        }

    def to_payload(self) -> RelationshipDictType:  # pylint: disable=missing-function-docstring
        payload: RelationshipDictType = {
            "dataset2Identifier": self.dataset2_identifier,
            "dataset1Keys": self.dataset1_keys,
            "dataset2Keys": self.dataset2_keys,
        }
        if self.dataset1_identifier:
            payload["dataset1Identifier"] = self.dataset1_identifier
        if self.feature_derivation_window_start is not None:
            payload["featureDerivationWindowStart"] = self.feature_derivation_window_start
        if self.feature_derivation_window_end is not None:
            payload["featureDerivationWindowEnd"] = self.feature_derivation_window_end
        if self.feature_derivation_window_time_unit:
            payload["featureDerivationWindowTimeUnit"] = self.feature_derivation_window_time_unit
        if self.feature_derivation_windows is not None:
            payload["featureDerivationWindows"] = self.feature_derivation_windows
        if self.prediction_point_rounding is not None:
            payload["predictionPointRounding"] = self.prediction_point_rounding
        if self.prediction_point_rounding_time_unit:
            payload["predictionPointRoundingTimeUnit"] = self.prediction_point_rounding_time_unit
        return payload


class SecondaryDataset:
    """A secondary dataset to be used for feature discovery

    .. versionadded:: v2.25

    Attributes
    ----------
    identifier: string
        Alias of the dataset (used directly as part of the generated feature names)
    catalog_id: string
        Identifier of the catalog item
    catalog_version_id: string
        Identifier of the catalog item version
    snapshot_policy: string, optional
        Policy to use while creating a project or making predictions.
        If omitted, by default endpoint will use 'latest'.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        dataset_definition = dr.SecondaryDataset(
            identifier='profile',
            catalog_id='5ec4aec1f072bc028e3471ae',
            catalog_version_id='5ec4aec2f072bc028e3471b1',
        )
    """

    def __init__(
        self,
        identifier: str,
        catalog_id: str,
        catalog_version_id: str,
        snapshot_policy: str = SNAPSHOT_POLICY.LATEST,
    ) -> None:
        self.identifier = identifier
        self.catalog_id = catalog_id
        self.catalog_version_id = catalog_version_id
        self.snapshot_policy = snapshot_policy

    def to_dict(self) -> SecondaryDatasetDict:
        return {
            "identifier": self.identifier,
            "catalog_id": self.catalog_id,
            "catalog_version_id": self.catalog_version_id,
            "snapshot_policy": self.snapshot_policy,
        }

    def to_payload(self) -> Dict[str, Optional[str]]:
        return {
            "identifier": self.identifier,
            "catalogId": self.catalog_id,
            "catalogVersionId": self.catalog_version_id,
            "snapshotPolicy": self.snapshot_policy,
        }


class FeatureDiscoverySetting:
    """A feature discovery settings used to customize the feature discovery process

    To see the list of possible settings, create a RelationshipConfiguration without specifying
    settings and check its `feature_discovery_settings` attribute, which is a list of possible
    settings with their default values.

    Attributes
    ----------
    name: str
        Name of the feature discovery setting
    value: bool
        Value of the feature discovery setting

    .. versionadded: v2.26
    """

    def __init__(self, name: str, value: bool) -> None:
        self.name = name
        self.value = value

    def to_dict(self) -> Dict[str, Union[str, bool]]:
        return {
            "name": self.name,
            "value": self.value,
        }



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import trafaret as t

from datarobot._compat import Int, String
from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.residuals import ResidualsTrafaret
from datarobot.utils.pagination import unpaginate

from .external_scores import DEFAULT_BATCH_SIZE


class ExternalResidualsChart(APIObject):
    """Residual analysis dataset chart data for model .
    This data is calculated over a randomly downsampled subset of the source data
    (capped at 1000 rows).

    versionadded:: v2.21

    Notes
    -----

    ``ResidualsChartRow`` is a list of floats and ints containing the following:
        * Element 0 (float) is the actual target value for the source data row.
        * Element 1 (float) is the predicted target value for that row.
        * Element 2 (float) is the error rate of predicted - actual and is optional.
        * Element 3 (int) is the row number in the source dataset from which the values
          were selected and is optional.

    Attributes
    ----------
    data : list
        List of lists with schema described as ``ResidualsChartRow`` above.
    coefficient_of_determination : float
        The r-squared value for the downsampled dataset
    residual_mean : float
        The arithmetic mean of the residual (predicted value minus actual value)
    dataset_id : str
        ID of the dataset this chart belongs
    standard_deviation : float
        standard deviation of residual values
    """

    _converter = (
        t.Dict(
            {
                t.Key("dataset_id"): String,
                t.Key("data"): t.List(t.Tuple(t.Float, t.Float, t.Float, t.Or(Int, t.Null))),
            }
        )
        .merge(ResidualsTrafaret)
        .ignore_extra("*")
    )

    _path = "projects/{}/models/{}/datasetResidualsCharts/"

    def __init__(
        self,
        dataset_id: str,
        residual_mean: float,
        coefficient_of_determination: float,
        standard_deviation: float,
        data: List[Tuple[float, float, float, Optional[int]]],
    ):
        self.dataset_id = dataset_id
        self.residual_mean = residual_mean
        self.coefficient_of_determination = coefficient_of_determination
        self.standard_deviation = standard_deviation
        self.data = data

    @classmethod
    def list(
        cls,
        project_id: str,
        model_id: str,
        dataset_id: Optional[str] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> List[ExternalResidualsChart]:
        """Retrieve list of residual charts for the model.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: str
            if specified, only lift chart for this model will be retrieved
        dataset_id: str, optional
            if specified, only lift chart for this dataset will be retrieved
        offset: int, optional
            this many results will be skipped, default: 0
        limit: int, optional
            at most this many results are returned, default: 100, max 1000.
            To return all results, specify 0

        Returns
        -------
            A list of :py:class:`ExternalResidualsChart <datarobot.ExternalResidualsChart>` objects
        """
        url = cls._path.format(project_id, model_id)
        params: Dict[str, Union[int, str]] = {"offset": offset, "limit": limit}
        if dataset_id:
            params["datasetId"] = dataset_id
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            return [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)]
        return [cls.from_server_data(i) for i in cls._client.get(url, params=params).json()["data"]]

    @classmethod
    def get(cls, project_id: str, model_id: str, dataset_id: str) -> ExternalResidualsChart:
        """Retrieve residual chart for the model and prediction dataset.

        Parameters
        ----------
        project_id: str
            project id
        model_id: str
            model id
        dataset_id: str
            prediction dataset id

        Returns
        -------
            :py:class:`ExternalResidualsChart <datarobot.ExternalResidualsChart>` object

        """
        resp = cls.list(project_id, model_id, dataset_id=dataset_id, offset=0, limit=1)
        if not resp:
            raise ClientError("Requested residual chart does not exist.", 404)
        return resp[0]

    def __repr__(self) -> str:
        return f"ExternalResidualChart({self.dataset_id})"



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional

import trafaret as t
from typing_extensions import TypedDict

from datarobot.enums import ApplicationPermissions
from datarobot.models.api_object import APIObject
from datarobot.models.use_cases.utils import resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate


class ApplicationDeployment(TypedDict):
    deployment_id: str
    reference_name: str


class ApplicationRelatedEntity(TypedDict):
    model_id: Optional[str]
    project_id: Optional[str]
    is_from_experiment_container: Optional[bool]


class Application(APIObject):
    """
    An entity associated with a DataRobot Application.

    Attributes
    ----------
    id : str
        The ID of the created application.
    application_type_id : str
        The ID of the type of the application.
    user_id : str
        The ID of the user which created the application.
    model_deployment_id : str
        The ID of the associated model deployment.
    deactivation_status_id : str or None
        The ID of the status object to track the asynchronous app deactivation process status.
        Will be None if the app was never deactivated.
    name : str
        The name of the application.
    created_by : str
        The username of the user created the application.
    created_at : str
        The timestamp when the application was created.
    updated_at : str
        The timestamp when the application was updated.
    datasets : List[str]
        The list of datasets IDs associated with the application.
    creator_first_name : Optional[str]
        Application creator first name. Optional.
    creator_last_name : Optional[str]
        Application creator last name. Optional.
    creator_userhash : Optional[str]
        Application creator userhash. Optional.
    deployment_status_id : str
        The ID of the status object to track the asynchronous deployment process status.
    description : str
        A description of the application.
    cloud_provider : str
        The host of this application.
    deployments : Optional[List[ApplicationDeployment]]
        A list of deployment details. Optional.
    deployment_ids : List[str]
        A list of deployment IDs for this app.
    deployment_name : Optional[str]
        Name of the deployment. Optional.
    application_template_type : Optional[str]
        Application template type, purpose. Optional.
    pool_used : bool
        Whether the pool where used for last app deployment.
    permissions : List[str]
        The list of permitted actions, which the authenticated user can perform on this application.
        Permissions should be ApplicationPermission options.
    has_custom_logo : bool
        Whether the app has a custom logo.
    related_entities : Optional[ApplcationRelatedEntity]
        IDs of entities, related to app for easy search.
    org_id : str
        ID of the app's organization.
    """

    _path = "applications/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("application_type_id"): t.String,
            t.Key("user_id"): t.String,
            t.Key("model_deployment_id"): t.String,
            t.Key("name"): t.String,
            t.Key("created_by"): t.String,
            t.Key("created_at"): t.String,
            t.Key("updated_at"): t.String,
            t.Key("datasets"): t.List(t.String),
            t.Key("deployment_status_id", optional=True): t.String,
            t.Key("description", optional=True): t.String,
            t.Key("cloud_provider"): t.String,
            t.Key("deployment_ids"): t.List(t.String),
            t.Key("deployment_name", optional=True): t.String,
            t.Key("permissions"): t.List(t.String),
            t.Key("has_custom_logo"): t.Bool,
            t.Key("org_id"): t.String,
            t.Key("pool_used"): t.Bool,
            t.Key("related_entities", optional=True): t.Dict(
                {
                    t.Key("model_id", optional=True): t.String,
                    t.Key("project_id", optional=True): t.String,
                    t.Key("is_from_experiment_container", optional=True): t.Bool,
                }
            ),
            t.Key("application_template_type", optional=True): t.String,
            t.Key("deactivation_status_id", optional=True): t.String,
            t.Key("created_first_name", optional=True): t.String,
            t.Key("creator_last_name", optional=True): t.String,
            t.Key("creator_userhash", optional=True): t.String,
            t.Key("deployments", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("deployment_id"): t.String,
                        t.Key("reference_name"): t.String,
                    }
                )
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        application_type_id: str,
        user_id: str,
        model_deployment_id: str,
        name: str,
        created_by: str,
        created_at: str,
        updated_at: str,
        datasets: List[str],
        cloud_provider: str,
        deployment_ids: List[str],
        pool_used: bool,
        permissions: List[str],
        has_custom_logo: bool,
        org_id: str,
        deployment_status_id: Optional[str] = None,
        description: Optional[str] = None,
        related_entities: Optional[ApplicationRelatedEntity] = None,
        application_template_type: Optional[str] = None,
        deployment_name: Optional[str] = None,
        deactivation_status_id: Optional[str] = None,
        created_first_name: Optional[str] = None,
        creator_last_name: Optional[str] = None,
        creator_userhash: Optional[str] = None,
        deployments: Optional[List[ApplicationDeployment]] = None,
    ):
        self.id = id
        self.application_type_id = application_type_id
        self.user_id = user_id
        self.model_deployment_id = model_deployment_id
        self.name = name
        self.created_by = created_by
        self.created_at = created_at
        self.updated_at = updated_at
        self.datasets = datasets
        self.deployment_status_id = deployment_status_id
        self.description = description
        self.cloud_provider = cloud_provider
        self.deployment_ids = deployment_ids
        self.pool_used = pool_used
        self.permissions = [ApplicationPermissions[permission] for permission in permissions]
        self.has_custom_logo = has_custom_logo
        self.org_id = org_id
        self.related_entities = related_entities
        self.application_template_type = application_template_type
        self.deployment_name = deployment_name
        self.deactivation_status_id = deactivation_status_id
        self.created_first_name = created_first_name
        self.creator_last_name = creator_last_name
        self.creator_userhash = creator_userhash
        self.deployments = deployments

    @classmethod
    def list(
        cls,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        use_cases: Optional[UseCaseLike] = None,
    ) -> List[Application]:
        """
        Retrieve a list of user applications.

        Parameters
        ----------
        offset : Optional[int]
            Optional. Retrieve applications in a list after this number.
        limit : Optional[int]
            Optional. Retrieve only this number of applications.
        use_cases: Optional[Union[UseCase, List[UseCase], str, List[str]]]
            Optional. Filter available Applications by a specific Use Case or Use Cases.
            Accepts either the entity or the ID.

        Returns
        -------
        applications : List[Application]
            The requested list of user applications.
        """
        query = {"offset": offset, "limit": limit}
        query = resolve_use_cases(use_cases=use_cases, params=query)
        applications = unpaginate(initial_url=cls._path, initial_params=query, client=cls._client)
        return [cls.from_server_data(application) for application in applications]

    @classmethod
    def get(cls, application_id: str) -> Application:
        """
        Retrieve a single application.

        Parameters
        ----------
        application_id : str
            The ID of the application to retrieve.

        Returns
        -------
        application : Application
            The requested application.
        """
        r_data = cls._client.get(f"{cls._path}{application_id}/")
        return cls.from_server_data(r_data.json())



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
import os
from typing import List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.utils.waiters import wait_for_async_resolution

from ..enums import DEFAULT_MAX_WAIT, DEFAULT_TIMEOUT
from ..utils import parse_time
from .api_object import APIObject


class ImportedModel(APIObject):
    """
    Represents an imported model available for making predictions. These are only relevant for
    administrators of on-premise Stand Alone Scoring Engines.

    ImportedModels are trained in one DataRobot application, exported as a `.drmodel` file, and
    then imported for use in a Stand Alone Scoring Engine.

    Attributes
    ----------
    id : str
        id of the import
    model_name : str
        model type describing the model generated by DataRobot
    display_name : str
        manually specified human-readable name of the imported model
    note : str
        manually added node about this imported model
    imported_at : datetime
        the time the model was imported
    imported_by_username : str
        username of the user who imported the model
    imported_by_id : str
        id of the user who imported the model
    origin_url : str
        URL of the application the model was exported from
    model_id : str
        original id of the model prior to export
    featurelist_name : str
        name of the featurelist used to train the model
    project_id : str
         id of the project the model belonged to prior to export
    project_name : str
        name of the project the model belonged to prior to export
    target : str
        the target of the project the model belonged to prior to export
    dataset_name : str
        filename of the dataset used to create the project the model belonged to
    created_by_username : str
        username of the user who created the model prior to export
    created_by_id : str
        id of the user who created the model prior to export
    """

    _root_path = "importedModels/"
    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("imported_at"): parse_time,
            t.Key("model_id"): String,
            t.Key("target"): String,
            t.Key("featurelist_name"): String,
            t.Key("dataset_name"): String,
            t.Key("model_name"): String,
            t.Key("project_id"): String,
            t.Key("note", optional=True, default=None): t.Or(
                String, t.Null
            ),  # from_api strips None
            t.Key("origin_url"): String,
            t.Key("imported_by_username"): String,
            t.Key("project_name"): String,
            t.Key("created_by_username"): String,
            t.Key("created_by_id"): String,
            t.Key("imported_by_id"): String,
            t.Key("display_name"): String,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        imported_at: Optional[datetime] = None,
        model_id: Optional[str] = None,
        target: Optional[str] = None,
        featurelist_name: Optional[str] = None,
        dataset_name: Optional[str] = None,
        model_name: Optional[str] = None,
        project_id: Optional[str] = None,
        note: Optional[str] = None,
        origin_url: Optional[str] = None,
        imported_by_username: Optional[str] = None,
        project_name: Optional[str] = None,
        created_by_username: Optional[str] = None,
        created_by_id: Optional[str] = None,
        imported_by_id: Optional[str] = None,
        display_name: Optional[str] = None,
    ) -> None:

        self.id = id
        self.imported_at = imported_at
        self.model_id = model_id
        self.target = target
        self.featurelist_name = featurelist_name
        self.dataset_name = dataset_name
        self.model_name = model_name
        self.project_id = project_id
        self.note = note
        self.origin_url = origin_url
        self.imported_by_username = imported_by_username
        self.project_name = project_name
        self.created_by_username = created_by_username
        self.created_by_id = created_by_id
        self.imported_by_id = imported_by_id
        self.display_name = display_name

        self._path = self._get_path(id)

    @classmethod
    def create(cls, path: str, max_wait: int = DEFAULT_MAX_WAIT) -> ImportedModel:
        """Import a previously exported model for predictions.

        Parameters
        ----------
        path : str
            The path to the exported model file
        max_wait : int, optional
            Time in seconds after which model import is considered unsuccessful
        """
        name = os.path.split(path)[1]
        response = cls._client.build_request_with_file(
            "post", cls._root_path, name, file_path=path, read_timeout=DEFAULT_TIMEOUT.UPLOAD
        )
        async_loc = response.headers["Location"]
        imported_model_loc = wait_for_async_resolution(cls._client, async_loc, max_wait=max_wait)
        return cls.from_location(imported_model_loc)

    @classmethod
    def _get_path(cls, id: str) -> str:
        return f"{cls._root_path}{id}/"

    @classmethod
    def get(cls, import_id: str) -> ImportedModel:
        """Retrieve imported model info

        Parameters
        ----------
        import_id : str
            The ID of the imported model.

        Returns
        -------
        imported_model : ImportedModel
            The ImportedModel instance
        """
        path = f"{cls._root_path}{import_id}/"
        return cls.from_location(path)

    @classmethod
    def list(cls, limit: Optional[int] = None, offset: Optional[int] = None) -> List[ImportedModel]:
        """
        List the imported models.

        Parameters
        ----------
        limit : int
            The number of records to return. The server will use a (possibly finite) default if not
            specified.
        offset : int
            The number of records to skip.


        Returns
        -------
        imported_models : list[ImportedModel]
        """
        r_data = cls._client.get(cls._root_path, params={"limit": limit, "offset": offset}).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    def update(self, display_name: Optional[str] = None, note: Optional[str] = None) -> None:
        """
        Update the display name or note for an imported model. The ImportedModel object is updated
        in place.

        Parameters
        ----------
        display_name : str
            The new display name.
        note : str
            The new note.
        """
        self._client.patch(self._path, data={"display_name": display_name, "note": note})
        if display_name:  # Following pattern from Project.update. (Arguably technical debt.)
            self.display_name = display_name
        if note:
            self.note = note

    def delete(self) -> None:
        """
        Delete this imported model.
        """
        self._client.delete(self._path)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from collections import OrderedDict
from datetime import datetime
from typing import cast, List, Optional, TYPE_CHECKING, Union

import dateutil
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import SERVICE_STAT_METRIC
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.utils import from_api

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class Period(TypedDict, total=False):
        """Type dict for period"""

        start: str
        end: str

    class Bucket(TypedDict, total=False):
        """Type dict for bucket"""

        period: Optional[Period]
        value: Optional[Union[int, float]]

    class Metrics(TypedDict, total=False):
        """Type dict for metrics"""

        totalPredictions: int
        totalRequests: int
        slowRequests: int
        executionTime: Optional[float]
        responseTime: Optional[float]
        userErrorRate: float
        serverErrorRate: float
        numConsumers: int
        cacheHitRatio: float
        medianLoad: float
        peakLoad: float


class ServiceStats(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment service stats information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve service stats metrics
    period : dict
        the time period used to retrieve service stats metrics
    metrics : dict
        the service stats metrics
    """

    _path = "deployments/{}/serviceStats/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> dateutil.parser.parse,
            t.Key("end"): t.String >> dateutil.parser.parse,
        }
    )
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metrics"): t.Dict(
                {
                    t.Key(SERVICE_STAT_METRIC.TOTAL_PREDICTIONS): t.Int(),
                    t.Key(SERVICE_STAT_METRIC.TOTAL_REQUESTS, optional=True): t.Int(),
                    t.Key(SERVICE_STAT_METRIC.SLOW_REQUESTS, optional=True): t.Int(),
                    t.Key(SERVICE_STAT_METRIC.EXECUTION_TIME, optional=True): t.Or(
                        t.Float(), t.Null()
                    ),
                    t.Key(SERVICE_STAT_METRIC.RESPONSE_TIME, optional=True): t.Or(
                        t.Float(), t.Null()
                    ),
                    t.Key(SERVICE_STAT_METRIC.USER_ERROR_RATE, optional=True): t.Float(),
                    t.Key(SERVICE_STAT_METRIC.SERVER_ERROR_RATE, optional=True): t.Float(),
                    t.Key(SERVICE_STAT_METRIC.NUM_CONSUMERS, optional=True): t.Int(),
                    t.Key(SERVICE_STAT_METRIC.CACHE_HIT_RATIO, optional=True): t.Float(),
                    t.Key(SERVICE_STAT_METRIC.MEDIAN_LOAD, optional=True): t.Float(),
                    t.Key(SERVICE_STAT_METRIC.PEAK_LOAD, optional=True): t.Float(),
                }
            ).allow_extra("*", trafaret=t.Or(t.Int(), t.Float(), t.Null())),
            t.Key("model_id"): t.Or(t.String(), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        period: Optional[Period] = None,
        metrics: Optional[Metrics] = None,
        model_id: Optional[str] = None,
    ) -> None:
        self.period = period or {}
        self.metrics = metrics or {}
        self.model_id = model_id

    def __repr__(self) -> str:
        return "{}({} | {} - {})".format(
            self.__class__.__name__,
            self.model_id,
            self.period.get("start"),
            self.period.get("end"),
        )

    def __getitem__(self, item: str) -> Optional[Union[int, float]]:
        return cast(Union[int, float], self.metrics.get(item))

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        execution_time_quantile: Optional[float] = None,
        response_time_quantile: Optional[float] = None,
        slow_requests_threshold: Optional[float] = None,
    ) -> ServiceStats:
        """Retrieve value of service stat metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        execution_time_quantile : float, optional
            quantile for `executionTime`, defaults to 0.5
        response_time_quantile : float, optional
            quantile for `responseTime`, defaults to 0.5
        slow_requests_threshold : float, optional
            threshold for `slowRequests`, defaults to 1000

        Returns
        -------
        service_stats : ServiceStats
            the queried service stats metrics
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            execution_time_quantile=execution_time_quantile,
            response_time_quantile=response_time_quantile,
            slow_requests_threshold=slow_requests_threshold,
        )
        data = cls._client.get(path, params=params).json()

        # we don't want to convert keys of the metrics object
        metrics = data.pop("metrics")

        data = from_api(data, keep_null_keys=True)
        data["metrics"] = metrics
        return cls.from_data(data)


class ServiceStatsOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment service stats over time information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve accuracy metric
    metric : str
        the service stat metric being retrieved
    buckets : dict
        how the service stat metric changes over time
    summary : dict
        summary for the service stat metric
    """

    _path = "deployments/{}/serviceStatsOverTime/"
    _period = t.Dict(
        {
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }
    )
    _bucket = t.Dict(
        {t.Key("period"): t.Or(_period, t.Null), t.Key("value"): t.Or(Int, t.Float, t.Null)}
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("buckets"): t.List(_bucket),
            t.Key("summary"): _bucket,
            t.Key("metric"): String(),
            t.Key("model_id"): t.Or(String(), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        buckets: Optional[List[Bucket]] = None,
        summary: Optional[Bucket] = None,
        metric: Optional[str] = None,
        model_id: Optional[str] = None,
    ) -> None:
        self.buckets = buckets or []
        self.summary = summary or {}
        self.metric = metric
        self.model_id = model_id

    def __repr__(self) -> str:
        period = self.summary.get("period") or {}
        return "{}({} | {} | {} - {})".format(
            self.__class__.__name__,
            self.model_id,
            self.metric,
            period.get("start"),
            period.get("end"),
        )

    @classmethod
    def get(
        cls,
        deployment_id: str,
        metric: Optional[str] = None,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[str] = None,
        quantile: Optional[float] = None,
        threshold: Optional[int] = None,
    ) -> ServiceStatsOverTime:
        """Retrieve information about how a service stat metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metric : SERVICE_STAT_METRIC, optional
            the service stat metric to retrieve
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        bucket_size : str, optional
            time duration of a bucket, in ISO 8601 time duration format
        quantile : float, optional
            quantile for 'executionTime' or 'responseTime', ignored when querying other metrics
        threshold : int, optional
            threshold for 'slowQueries', ignored when querying other metrics

        Returns
        -------
        service_stats_over_time : ServiceStatsOverTime
            the queried service stat over time information
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            metric=metric,
            bucket_size=bucket_size,
            quantile=quantile,
            threshold=threshold,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)

    @property
    def bucket_values(self) -> OrderedDict[str, Union[int, float, None]]:
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: OrderedDict
        """

        values = [
            (bucket["period"]["start"], bucket["value"])
            for bucket in self.buckets
            if bucket["period"]
        ]
        return OrderedDict(values)



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from setuptools import find_packages, setup

from common_setup import common_setup_kwargs, DEFAULT_CLASSIFIERS, DESCRIPTION_TEMPLATE, version

if "b" in version:
    msg = (
        "Major releases must not have a 'b' for beta in the version. "
        "Go back and make a release branch with tag, then update the version number."
    )
    raise RuntimeError(msg)

python_versions = ">= 3.7"

description = DESCRIPTION_TEMPLATE.format(
    package_name="datarobot",
    pypi_url_target="https://pypi.python.org/pypi/datarobot/",
    extra_desc="",
    python_versions=python_versions,
    pip_package_name="datarobot",
    docs_link="https://datarobot-public-api-client.readthedocs-hosted.com",
)

packages = find_packages(exclude=["tests*", "*_experimental*"])

common_setup_kwargs.update(
    name="datarobot",
    version=version,
    packages=packages,
    long_description=description,
    classifiers=DEFAULT_CLASSIFIERS,
)

setup(**common_setup_kwargs)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Iterable, List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.enums import CUSTOM_TASK_LANGUAGE, CUSTOM_TASK_TARGET_TYPE
from datarobot.models.api_object import APIObject, ServerDataType
from datarobot.models.custom_task_version import CustomTaskVersion
from datarobot.models.sharing import SharingAccess
from datarobot.utils.pagination import unpaginate


class CustomTask(APIObject):
    """A custom task. This can be in a partial state or a complete state.
    When the `latest_version` is `None`, the empty task has been initialized with
    some metadata.  It is not yet use-able for actual training.  Once the first
    `CustomTaskVersion` has been created, you can put the CustomTask in UserBlueprints to
    train Models in Projects

    .. versionadded:: v2.26

    Attributes
    ----------
    id: str
        id of the custom task
    name: str
        name of the custom task
    language: str
        programming language of the custom task.
        Can be "python", "r", "java" or "other"
    description: str
        description of the custom task
    target_type: datarobot.enums.CUSTOM_TASK_TARGET_TYPE
        the target type of the custom task. One of:

        - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.BINARY`
        - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.REGRESSION`
        - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.MULTICLASS`
        - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.ANOMALY`
        - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.TRANSFORM`
    latest_version: datarobot.CustomTaskVersion or None
        latest version of the custom task if the task has a latest version. If the
        latest version is None, the custom task is not ready for use in user blueprints.
        You must create its first CustomTaskVersion before you can use the CustomTask
    created_by: str
        The username of the user who created the custom task.
    updated_at: str
        An ISO-8601 formatted timestamp of when the custom task was updated.
    created_at: str
        ISO-8601 formatted timestamp of when the custom task was created
    calibrate_predictions: bool
        whether anomaly predictions should be calibrated to be between 0 and 1 by DR.
        only applies to custom estimators with target type
        `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.ANOMALY`
    """

    _path = "customTasks/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("target_type"): String(),
            t.Key("latest_version", optional=True, default=None): CustomTaskVersion.schema
            | t.Null(),
            t.Key("created") >> "created_at": String(),
            t.Key("updated") >> "updated_at": String(),
            t.Key("name"): String(),
            t.Key("description"): String(allow_blank=True),
            t.Key("language"): String(allow_blank=True),
            t.Key("created_by"): String(),
            t.Key("calibrate_predictions", optional=True): t.Bool(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        target_type: CUSTOM_TASK_TARGET_TYPE,
        latest_version: Optional[CustomTaskVersion],
        created_at: str,
        updated_at: str,
        name: str,
        description: str,
        language: CUSTOM_TASK_LANGUAGE,
        created_by: str,
        calibrate_predictions: Optional[bool] = None,
    ) -> None:
        if latest_version is not None:
            # TODO: Annotate CustomTaskVersion [DSX-2323]
            latest_version = CustomTaskVersion(**latest_version)  # type: ignore[no-untyped-call, arg-type]

        self.id = id
        self.target_type = target_type
        self.latest_version = latest_version
        self.created_at = created_at
        self.updated_at = updated_at
        self.name = name
        self.description = description
        self.language = language
        self.created_by = created_by
        self.calibrate_predictions = calibrate_predictions

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _update_values(self, new_response: CustomTask) -> None:
        # type (CustomTask) -> None
        for attr in self._fields():  # type: ignore[no-untyped-call]
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def from_server_data(
        cls, data: ServerDataType, keep_attrs: Optional[Iterable[str]] = None
    ) -> CustomTask:
        raw_task: CustomTask = super().from_server_data(data, keep_attrs)
        # from_server_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        latest_version_data = data.get("latestVersion")  # type: ignore[union-attr]
        if latest_version_data is not None:
            raw_task.latest_version.required_metadata = latest_version_data.get(  # type: ignore[union-attr]
                "requiredMetadata"
            )
        return raw_task

    @classmethod
    def list(
        cls, order_by: Optional[str] = None, search_for: Optional[str] = None
    ) -> List[CustomTask]:
        """List custom tasks available to the user.

        .. versionadded:: v2.26

        Parameters
        ----------
        search_for: str, optional
            string for filtering custom tasks - only tasks that contain the
            string in name or description will be returned.
            If not specified, all custom task will be returned
        order_by: str, optional
            property to sort custom tasks by.
            Supported properties are "created" and "updated".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None which will result in
            custom tasks being returned in order of creation time descending

        Returns
        -------
        List[CustomTask]
            a list of custom tasks.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "order_by": order_by,
            "search_for": search_for,
        }
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_task_id: str) -> CustomTask:
        """Get custom task by id.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            id of the custom task

        Returns
        -------
        CustomTask
            retrieved custom task

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = f"{cls._path}{custom_task_id}/"
        data = cls._client.get(path).json()
        return cls.from_server_data(data)

    @classmethod
    def copy(cls, custom_task_id: str) -> CustomTask:
        """Create a custom task by copying existing one.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            id of the custom task to copy

        Returns
        -------
        CustomTask

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{cls._path}fromCustomTask/"
        response = cls._client.post(path, data={"custom_task_id": custom_task_id})
        return cls.from_server_data(response.json())

    @classmethod
    def create(  # type: ignore[no-untyped-def]
        cls,
        name: str,
        target_type: CUSTOM_TASK_TARGET_TYPE,
        language: Optional[CUSTOM_TASK_LANGUAGE] = None,
        description: Optional[str] = None,
        calibrate_predictions: Optional[bool] = None,
        **kwargs,
    ) -> CustomTask:
        """
        Creates *only the metadata* for a custom task.  This task will
        not be use-able until you have created a CustomTaskVersion attached to this task.

        .. versionadded:: v2.26

        Parameters
        ----------
        name: str
            name of the custom task
        target_type: datarobot.enums.CUSTOM_TASK_TARGET_TYPE
            the target typed based on the following values. Anything else will raise an error

            - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.BINARY`
            - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.REGRESSION`
            - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.MULTICLASS`
            - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.ANOMALY`
            - `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.TRANSFORM`
        language: str, optional
            programming language of the custom task.
            Can be "python", "r", "java" or "other"
        description: str, optional
            description of the custom task
        calibrate_predictions: bool, optional
            whether anomaly predictions should be calibrated to be between 0 and 1 by DR.
            if None, uses default value from DR app (True).
            only applies to custom estimators with target type
            `datarobot.enums.CUSTOM_TASK_TARGET_TYPE.ANOMALY`

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.

        Returns
        -------
        CustomTask
        """
        cls._validate_target_type(target_type)
        payload = {k: v for k, v in kwargs.items()}  # pylint: disable=unnecessary-comprehension
        payload.update({"name": name, "target_type": target_type})
        for k, v in [
            ("language", language),
            ("description", description),
            ("calibrate_predictions", calibrate_predictions),
        ]:
            if v is not None:
                payload[k] = v

        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def _validate_target_type(cls, target_type: CUSTOM_TASK_TARGET_TYPE) -> None:
        if target_type not in CUSTOM_TASK_TARGET_TYPE.ALL:
            raise ValueError(f"{target_type} is not one of {CUSTOM_TASK_TARGET_TYPE.ALL}")

    def update(  # type: ignore[no-untyped-def]
        self,
        name: Optional[str] = None,
        language: Optional[CUSTOM_TASK_LANGUAGE] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Update custom task properties.

        .. versionadded:: v2.26

        Parameters
        ----------
        name: str, optional
            new custom task name
        language: str, optional
            new custom task programming language
        description: str, optional
            new custom task description

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {k: v for k, v in kwargs.items()}  # pylint: disable=unnecessary-comprehension
        for k, v in [("name", name), ("language", language), ("description", description)]:
            if v is not None:
                payload[k] = v

        url = f"{self._path}{self.id}/"
        data = self._client.patch(url, data=payload).json()
        new_obj = self.from_server_data(data)
        self._update_values(new_obj)

    def refresh(self) -> None:
        """Update custom task with the latest data from server.

        .. versionadded:: v2.26

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.id)
        self._update_values(new_object)

    def delete(self) -> None:
        """Delete custom task.

        .. versionadded:: v2.26

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = f"{self._path}{self.id}/"
        self._client.delete(url)

    def download_latest_version(self, file_path: str) -> None:
        """Download the latest custom task version.

        .. versionadded:: v2.26

        Parameters
        ----------
        file_path: str
            the full path of the target zip file

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = f"{self._path}{self.id}/download/"
        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def get_access_list(self) -> List[SharingAccess]:
        """Retrieve access control settings of this custom task.

        .. versionadded:: v2.27

        Returns
        -------
        list of :class:`SharingAccess <datarobot.SharingAccess>`
        """
        url = f"{self._path}{self.id}/accessControl/"
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, {}, self._client)
        ]

    def share(self, access_list: List[SharingAccess]) -> None:
        """Update the access control settings of this custom task.

        .. versionadded:: v2.27

        Parameters
        ----------
        access_list : list of :class:`SharingAccess <datarobot.SharingAccess>`
            A list of SharingAccess to update.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Examples
        --------
        Transfer access to the custom task from old_user@datarobot.com to new_user@datarobot.com

        .. code-block:: python

            import datarobot as dr

            new_access = dr.SharingAccess(new_user@datarobot.com,
                                          dr.enums.SHARING_ROLE.OWNER, can_share=True)
            access_list = [dr.SharingAccess(old_user@datarobot.com, None), new_access]

            dr.CustomTask.get('custom-task-id').share(access_list)
        """
        payload = {
            "data": [access.collect_payload() for access in access_list],
        }
        nullable_query_params = {"role"}
        self._client.patch(
            f"{self._path}{self.id}/accessControl/",
            data=payload,
            keep_attrs=nullable_query_params,
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import time
from typing import Any, Callable, TYPE_CHECKING

from datarobot import errors

if TYPE_CHECKING:
    from requests import Response

    from datarobot.rest import RESTClientObject


def wait_for_custom_resolution(
    client: RESTClientObject,
    url: str,
    success_fn: Callable[[Response], Any],
    max_wait: int = 600,
) -> Any:
    """
    Poll a url until success_fn returns something truthy

    Parameters
    ----------
    client : RESTClientObject
        The configured v2 requests session
    url : str
        The URL we are polling for resolution. This can be either a fully-qualified URL
        like `http://host.com/routeName/` or just the relative route within the API
        i.e. `routeName/`.
    success_fn : Callable[[requests.Response], Any]
        The method to determine if polling should finish. If the method returns a truthy value,
        polling will stop and this value will be returned.
    max_wait : int
        The number of seconds to wait before giving up

    Returns
    -------
    Any
        The final value returned by success_fn

    Raises
    ------
    AsyncFailureError
        If any of the responses from the server are unexpected
    AsyncTimeoutError
        If the resource did not resolve in time
    """
    start_time = time.time()

    join_endpoint = not url.startswith("http")  # Accept full qualified and relative urls

    response = client.get(url, allow_redirects=False, join_endpoint=join_endpoint)
    while time.time() < start_time + max_wait:
        if response.status_code not in (200, 303, 307):
            e_template = "The server gave an unexpected response. Status Code {}: {}"
            raise errors.AsyncFailureError(e_template.format(response.status_code, response.text))
        is_successful = success_fn(response)

        if is_successful:
            return is_successful

        time.sleep(5)
        response = client.get(url, allow_redirects=False, join_endpoint=join_endpoint)

    timeout_msg = "Client timed out in {} seconds waiting for {} to resolve. Last status was {}: {}"
    raise errors.AsyncTimeoutError(
        timeout_msg.format(max_wait, url, response.status_code, response.text)
    )


def wait_for_async_resolution(
    client: RESTClientObject,
    async_location: str,
    max_wait: int = 600,
) -> Any:
    """
    Wait for successful resolution of the provided async_location.

    Parameters
    ----------
    client : RESTClientObject
        The configured v2 requests session
    async_location : str
        The URL we are polling for resolution. This can be either a fully-qualified URL
        like `http://host.com/routeName/` or just the relative route within the API
        i.e. `routeName/`.
    max_wait : int
        The number of seconds to wait before giving up

    Returns
    -------
    Any
        The final value returned by success_fn
        Can be the URL of the now-finished resource

    Raises
    ------
    AsyncFailureError
        If any of the responses from the server are unexpected
    AsyncProcessUnsuccessfulError
        If the job being waited for has failed or has been cancelled.
    AsyncTimeoutError
        If the resource did not resolve in time
    """

    def async_resolved(response: Response) -> Any:
        if response.status_code == 307:
            response = client.get(response.headers["Location"], allow_redirects=False)
        if response.status_code == 303:
            return response.headers["Location"]
        data = response.json()
        if data["status"].lower()[:5] in ["error", "abort"]:
            e_template = "The job did not complete successfully. Job Data: {}"
            raise errors.AsyncProcessUnsuccessfulError(e_template.format(data))
        if data["status"].lower() == "completed":
            return data

    return wait_for_custom_resolution(client, async_location, async_resolved, max_wait)



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
__version__ = "3.2.1"
__expected_server_version__ = "2.31"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import itertools
import time
from typing import Generator, Optional, Tuple, Union


def wait(
    timeout: Optional[Union[int, float]],
    delay: Union[int, float] = 0.1,
    maxdelay: Union[int, float] = 1.0,
) -> Generator[Tuple[int, float], None, None]:
    """Generate a slow loop, with exponential back-off.

    Parameters
    ----------
    timeout : float or int
        Total seconds to wait.
    delay : float or int
        Initial seconds to sleep.
    maxdelay : float or int
        Maximum seconds to sleep.

    Yields
    ----------
    int
        Retry count.

    Examples
    ----------
    >>> for index in retry.wait(10):
        # break if condition is met
    """
    if timeout is None:
        timeout = float("Inf")
    start_time = time.time()
    delay /= 2.0
    for index in itertools.count():
        seconds_waited = time.time() - start_time
        remaining = timeout - seconds_waited
        yield index, seconds_waited
        if remaining < 0:
            break
        delay = min(delay * 2, maxdelay, remaining)
        time.sleep(delay)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.models.execution_environment_version import ExecutionEnvironmentVersion
from datarobot.utils.pagination import unpaginate


class RequiredMetadataKey(APIObject):
    """Definition of a metadata key that custom models using this environment must define

    .. versionadded:: v2.25

    Attributes
    ----------
    field_name: str
        The required field key. This value will be added as an environment
        variable when running custom models.
    display_name: str
        A human readable name for the required field.
    """

    _converter = t.Dict({t.Key("field_name"): String(), t.Key("display_name"): String()})

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return "{}(field_name={!r}, display_name={!r})".format(
            self.__class__.__name__,
            self.field_name,
            self.display_name,
        )

    def _set_values(self, field_name, display_name):
        self.field_name = field_name
        self.display_name = display_name

    def to_dict(self):
        return self._converter.check(
            {"field_name": self.field_name, "display_name": self.display_name}
        )


class ExecutionEnvironment(APIObject):
    """An execution environment entity.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        the id of the execution environment
    name: str
        the name of the execution environment
    description: str, optional
        the description of the execution environment
    programming_language: str, optional
        the programming language of the execution environment.
        Can be "python", "r", "java" or "other"
    is_public: bool, optional
        public accessibility of environment, visible only for admin user
    created_at: str, optional
        ISO-8601 formatted timestamp of when the execution environment version was created
    latest_version: ExecutionEnvironmentVersion, optional
        the latest version of the execution environment
    """

    _path = "executionEnvironments/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(max_length=255),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("programming_language", optional=True): String(),
            t.Key("is_public", optional=True): t.Bool(),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("latest_version", optional=True): ExecutionEnvironmentVersion.schema,
            t.Key("required_metadata_keys", optional=True, default=list): t.List(
                RequiredMetadataKey.schema
            ),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id,
        name,
        description=None,
        programming_language=None,
        is_public=None,
        created_at=None,
        latest_version=None,
        required_metadata_keys=None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.programming_language = programming_language
        self.is_public = is_public
        self.created_at = created_at
        self.required_metadata_keys = [
            RequiredMetadataKey.from_data(key) for key in required_metadata_keys
        ]

        if latest_version is not None:
            latest_version.pop("image_id", None)  # "image_id" is being removed in RAPTOR-2460
            self.latest_version = ExecutionEnvironmentVersion(**latest_version)
        else:
            self.latest_version = None

    @classmethod
    def create(cls, name, description=None, programming_language=None, required_metadata_keys=None):
        """Create an execution environment.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str
            execution environment name
        description: str, optional
            execution environment description
        programming_language: str, optional
            programming language of the environment to be created.
            Can be "python", "r", "java" or "other". Default value - "other"
        required_metadata_keys: List[RequiredMetadataKey]
            Definition of a metadata keys that custom models using this environment must define

        Returns
        -------
        ExecutionEnvironment
            created execution environment

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        required_metadata_keys = required_metadata_keys or []
        payload = {
            "name": name,
            "description": description,
            "programming_language": programming_language,
        }
        if required_metadata_keys:
            payload["required_metadata_keys"] = [key.to_dict() for key in required_metadata_keys]
        response = cls._client.post(cls._path, data=payload)
        environment_id = response.json()["id"]
        return cls.get(environment_id)

    @classmethod
    def list(cls, search_for=None):
        """List execution environments available to the user.

        .. versionadded:: v2.21

        Parameters
        ----------
        search_for: str, optional
            the string for filtering execution environment - only execution
            environments that contain the string in name or description will
            be returned.

        Returns
        -------
        List[ExecutionEnvironment]
            a list of execution environments.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(cls._path, {"search_for": search_for}, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, execution_environment_id):
        """Get execution environment by it's id.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            ID of the execution environment to retrieve

        Returns
        -------
        ExecutionEnvironment
            retrieved execution environment

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{cls._path}{execution_environment_id}/"
        return cls.from_location(path)

    def delete(self):
        """Delete execution environment.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = f"{self._path}{self.id}/"
        self._client.delete(url)

    def update(self, name=None, description=None, required_metadata_keys=None):
        """Update execution environment properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str, optional
            new execution environment name
        description: str, optional
            new execution environment description
        required_metadata_keys: List[RequiredMetadataKey]
            Definition of a metadata keys that custom models using this environment must define

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {"name": name, "description": description}
        if required_metadata_keys is not None:
            payload["required_metadata_keys"] = [key.to_dict() for key in required_metadata_keys]

        url = f"{self._path}{self.id}/"
        response = self._client.patch(url, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def refresh(self):
        """Update execution environment with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = f"{self._path}{self.id}/"
        response = self._client.get(url)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Optional

from .job import Job
from .shap_matrix import ShapMatrix


class ShapMatrixJob(Job):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        data: Dict[str, Any],
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(data, **kwargs)
        self._model_id = model_id
        self._dataset_id = dataset_id

    @classmethod
    def get(
        cls,
        project_id: str,
        job_id: str,
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
    ) -> ShapMatrixJob:
        """
        Fetches one SHAP matrix job.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job identifier
        model_id : str
            The identifier of the model used for computing prediction explanations
        dataset_id : str
            The identifier of the dataset against which prediction explanations should be computed

        Returns
        -------
        job : ShapMatrixJob
            The job

        Raises
        ------
            AsyncFailureError
                Querying this resource gave a status code other than 200 or 303
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(
            data,
            model_id=model_id,
            dataset_id=dataset_id,
            completed_resource_url=completed_url,
        )

    def _make_result_from_location(self, location: str, params: Optional[Any] = None) -> ShapMatrix:
        return ShapMatrix.from_location(
            location,
            model_id=self._model_id,
            dataset_id=self._dataset_id,
        )

    def refresh(self) -> None:
        """
        Update this object with the latest job data from the server.
        """
        data, completed_url = self._data_and_completed_url_for_job(self._this_job_path())
        # pylint: disable-next=unnecessary-dunder-call
        self.__init__(  # type: ignore[misc]
            data,
            model_id=self._model_id,
            dataset_id=self._dataset_id,
            completed_resource_url=completed_url,
        )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
#  pylint: disable=C0415
from __future__ import annotations

from dataclasses import dataclass
from types import TracebackType
from typing import Dict, List, Optional, Tuple, Type, TypeVar, Union

import trafaret as t

from datarobot.context import Context, DefaultUseCase
from datarobot.enums import (
    SHARING_RECIPIENT_TYPE,
    SHARING_ROLE,
    UseCaseAPIPathEntityType,
    UseCaseEntityType,
    UseCaseReferenceEntityMap,
)
from datarobot.errors import InvalidUsageError
from datarobot.models.api_object import APIObject
from datarobot.models.sharing import SharingRole
from datarobot.utils.pagination import unpaginate

T = TypeVar("T")

use_case_user_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("full_name", optional=True): t.Or(t.String, t.Null),
        t.Key("email", optional=True): t.Or(t.String, t.Null),
        t.Key("userhash", optional=True): t.Or(t.String, t.Null),
        t.Key("username", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


@dataclass
class UseCaseUser(APIObject):
    """Representation of a Use Case user.

    Attributes
    ----------
    id : str
        The id of the user.
    full_name : str
        The full name of the user. Optional.
    email : str
        The email address of the user. Optional.
    userhash : str
        User's gravatar hash. Optional.
    username : str
        The username of the user. Optional.
    """

    _converter = use_case_user_trafaret

    id: str
    full_name: Optional[str] = None
    email: Optional[str] = None
    userhash: Optional[str] = None
    username: Optional[str] = None


class UseCaseReferenceEntity(APIObject):
    """
    An entity associated with a Use Case.

    Attributes
    ----------
    entity_type : UseCaseEntityType
        The type of the entity.
    use_case_id : str
        The Use Case this entity is associated with.
    id : str
        The ID of the entity.
    created_at : str
        The date and time this entity was linked with the Use Case.
    is_deleted : bool
        Whether or not the linked entity has been deleted.
    created : UseCaseUser
        The user who created the link between this entity and the Use Case.
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("entity_type"): t.String,
            t.Key("entity_id"): t.String,
            t.Key("experiment_container_id") >> "use_case_id": t.String,
            t.Key("created_at"): t.String,
            t.Key("created"): use_case_user_trafaret,
            t.Key("is_deleted"): t.Bool,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        entity_type: UseCaseEntityType,
        entity_id: str,
        use_case_id: str,
        created_at: str,
        created: Dict[str, str],
        is_deleted: bool,
    ):
        self.id = id
        self.entity_type = entity_type
        self.entity_id = entity_id
        self.use_case_id = use_case_id
        self.created_at = created_at
        self.created = UseCaseUser(**created)
        self.is_deleted = is_deleted


def get_reference_entity_info(
    entity: Union[UseCaseReferenceEntity, T]
) -> Tuple[UseCaseAPIPathEntityType, Optional[str]]:
    """
    Get the entity type and entity id for a reference entity instance
    Parameters
    ----------
    entity : Union[UseCaseReferenceEntity, Project, Dataset, Application]
        The entity instance to add to a Use Case.
    Returns
    -------
    entity_info : Tuple[UseCaseEntityType, str]
        The entity type and entity id
    """
    from ..application import Application
    from ..dataset import Dataset
    from ..prediction_dataset import PredictionDataset
    from ..project import Project

    if isinstance(entity, Project):
        return UseCaseAPIPathEntityType.PROJECT, entity.id
    elif isinstance(entity, Dataset):
        return UseCaseAPIPathEntityType.DATASET, entity.id
    elif isinstance(entity, Application):
        return UseCaseAPIPathEntityType.APPLICATION, entity.id
    elif isinstance(entity, UseCaseReferenceEntity):
        entity_type = UseCaseReferenceEntityMap.get(entity.entity_type)
        if entity_type:
            return entity_type, entity.entity_id
    error_message = (
        f"Entity must be Project, Dataset, Application, "
        f"or UseCaseReferenceEntity. Invalid type: {type(entity).__name__}."
    )
    if isinstance(entity, UseCaseReferenceEntity):
        # If we're failing here, it's because the UseCaseReferenceEntity is a notebook or other unsupported type
        error_message = error_message + f" Unsupported Entity Type: {entity.entity_type}."
    if isinstance(entity, PredictionDataset):
        # Adding a specific error for PredictionDataset since the name would indicate it could
        # be treated as a Dataset, when in fact it's a separate entity and not a subclass.
        error_message = (
            error_message + " PredictionDataset is not a subclass of Dataset and "
            "cannot be added to a UseCase at this time."
        )
    raise TypeError(error_message)


class UseCase(APIObject):
    """Representation of a Use Case.

    Attributes
    ----------
    id : str
        The ID of the Use Case.
    name : str
        The name of the Use Case.
    description : str
        The description of the Use Case. Nullable.
    created_at  : str
        The timestamp generated at record creation.
    created : UseCaseUser
        The user who created the Use Case.
    updated_at : str
        The timestamp generated when the record was last updated.
    updated : UseCaseUser
        The most recent user to update the Use Case.
    models_count : int
        The number of models in a Use Case.
    projects_count : int
        The number of projects in a Use Case.
    datasets_count: int
        The number of datasets in a Use Case.
    notebooks_count: int
        The number of notebooks in a Use Case.
    applications_count: int
        The number of applications in a Use Case.
    owners : List[UseCaseUser]
        The most recent user to update the Use Case.
    members : List[UseCaseUser]
        The most recent user to update the Use Case.

    Examples
    --------

    .. code-block:: python

        import datarobot
        with UseCase.get("2348ac"):
            print(f"The current use case is {dr.Context.use_case}")
    """

    _path = "experimentContainers/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("name"): t.String,
            t.Key("description", optional=True): t.Or(t.String, t.Null),
            t.Key("created_at"): t.String,
            t.Key("created"): use_case_user_trafaret,
            t.Key("updated_at"): t.String,
            t.Key("updated"): use_case_user_trafaret,
            t.Key("models_count"): t.Int,
            t.Key("projects_count"): t.Int,
            t.Key("datasets_count"): t.Int,
            t.Key("notebooks_count"): t.Int,
            t.Key("applications_count"): t.Int,
            t.Key("owners", optional=True): t.List(use_case_user_trafaret),
            t.Key("members"): t.List(use_case_user_trafaret),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        created_at: str,
        created: Dict[str, str],
        updated_at: str,
        updated: Dict[str, str],
        models_count: int,
        projects_count: int,
        datasets_count: int,
        notebooks_count: int,
        applications_count: int,
        members: List[Dict[str, str]],
        description: Optional[str] = None,
        owners: Optional[List[Dict[str, str]]] = None,
    ):
        self.id: str = id
        self.name: str = name
        self.description: Optional[str] = description
        self.created_at: str = created_at
        self.created: UseCaseUser = UseCaseUser(**created)
        self.updated_at: str = updated_at
        self.updated: UseCaseUser = UseCaseUser(**updated)
        self.models_count: int = models_count
        self.projects_count: int = projects_count
        self.datasets_count: int = datasets_count
        self.notebooks_count: int = notebooks_count
        self.applications_count: int = applications_count
        self.members: List[UseCaseUser] = [UseCaseUser(**member) for member in members]
        self.owners: Optional[List[UseCaseUser]] = (
            [UseCaseUser(**owner) for owner in owners] if owners else None
        )

    def __repr__(self) -> str:
        return (
            f"UseCase(id={self.id}, "
            f"name={self.name}, "
            f"description={self.description}, "
            f"models={self.models_count}, "
            f"projects={self.projects_count}, "
            f"datasets={self.datasets_count}, "
            f"notebooks={self.notebooks_count}, "
            f"applications={self.applications_count})"
        )

    @classmethod
    def get(cls, use_case_id: str) -> UseCase:
        """
        Gets information about a Use Case.

        Parameters
        ----------
        use_case_id : str
            The identifier of the Use Case you want to load.

        Returns
        -------
        use_case : UseCase
            The queried Use Case.
        """
        path = f"{cls._path}{use_case_id}/"
        return cls.from_location(path)

    @classmethod
    def list(cls, search_params: Optional[Dict[str, Union[str, int]]] = None) -> List[UseCase]:
        """
        Returns the Use Cases associated with this account.

        Parameters
        ----------
        search_params : dict, optional.
            If not `None`, the returned projects are filtered by lookup.
            Currently, you can query use cases by:

            * ``offset`` - The number of records to skip over. Default 0.
            * ``limit`` - The number of records to return in the range from 1 to 100. Default 100.
            * ``search`` - Only return Use Cases with names that match the given string.
            * ``project_id`` - Only return Use Cases associated with the given project ID.
            * ``application_id`` - Only return Use Cases associated with the given app.
            * ``sort`` - The order to sort the Use Cases.

            ``sort`` queries can use the following options:

            * ``id`` or ``-id``
            * ``name`` or ``-name``
            * ``description`` or ``-description``
            * ``projects_count`` or ``-projects_count``
            * ``datasets_count`` or ``-datasets_count``
            * ``notebooks_count`` or ``-notebooks_count``
            * ``applications_count`` or ``-applications_count``
            * ``created_at`` or ``-created_at``
            * ``created_by`` or ``-created_by``
            * ``updated_at`` or ``-updated_at``
            * ``updated_by`` or ``-updated_by``

        Returns
        -------
        use_cases : list of UseCase instances
            Contains a list of Use Cases associated with this user
            account.

        Raises
        ------
        TypeError
            Raised if ``search_params`` parameter is provided,
            but is not of supported type.
        """
        get_params = {}
        if search_params is not None:
            if isinstance(search_params, dict):
                get_params.update(search_params)
            else:
                raise TypeError(
                    "Provided search_params argument {} is invalid type {}".format(
                        search_params, type(search_params)
                    )
                )
        r_data = unpaginate(cls._path, get_params, cls._client)
        return [cls.from_server_data(item) for item in r_data]

    @classmethod
    def create(cls, name: Optional[str] = None, description: Optional[str] = None) -> UseCase:
        """
        Create a new Use Case.

        Parameters
        ----------
        name : str
            Optional. The name of the new Use Case.
        description: str
            The description of the new Use Case. Optional.

        Returns
        -------
        use_case : UseCase
            The created Use Case.
        """
        payload = {"name": name, "description": description}
        use_case_id = cls._client.post(cls._path, data=payload).json()["id"]
        return cls.get(use_case_id)

    @classmethod
    def delete(cls, use_case_id: str) -> None:
        """
        Delete a Use Case.

        Parameters
        ----------
        use_case_id : str
            The ID of the Use Case to be deleted.
        """
        path = f"{cls._path}{use_case_id}/"
        cls._client.delete(path)
        return None

    @classmethod
    def _resolve_api_entity(
        cls, entity_type: Optional[UseCaseEntityType] = None
    ) -> UseCaseAPIPathEntityType | None:
        """
        For a given reference entity type, return the corresponding API endpoint type.
        """
        if entity_type == UseCaseEntityType.NOTEBOOK:
            raise InvalidUsageError(
                "Notebooks are currently unavailable for adding or removing from Use Cases via the API."
            )
        return UseCaseReferenceEntityMap.get(entity_type)

    def __enter__(self) -> UseCase:
        # Not declared in __init__ because this attribute has a very
        # specific use in context management and nowhere else
        # pylint: disable-next=attribute-defined-outside-init
        self.__previous: DefaultUseCase = Context.get_use_case(raw=True)

        Context.use_case = self
        return self

    def __exit__(
        self,
        __exc_type: Type[BaseException] | None,
        __exc_value: BaseException | None,
        __traceback: TracebackType | None,
    ) -> bool | None:
        Context.use_case = self.__previous
        return None

    def update(self, name: Optional[str] = None, description: Optional[str] = None) -> UseCase:
        """
        Update a Use Case's name or description.

        Parameters
        ----------
        name : str
            The updated name of the Use Case.
        description : str
            The updated description of the Use Case.

        Returns
        -------
        use_case : UseCase
            The updated Use Case.
        """
        payload = {"name": name, "description": description}
        path = f"{self._path}{self.id}/"
        r_data = self._client.patch(path, data=payload).json()
        return self.from_server_data(r_data)

    def add(
        self,
        entity: Optional[Union[UseCaseReferenceEntity, T]] = None,
        entity_type: Optional[UseCaseEntityType] = None,
        entity_id: Optional[str] = None,
    ) -> UseCaseReferenceEntity:
        """
        Add an entity (project, dataset, etc.) to a Use Case. Can only accept either an entity or
        an entity type and entity ID, but not both.

        Projects and Applications can only be linked to a single Use Case. Datasets can be linked to multiple Use Cases.

        There are some prerequisites for linking Projects to a Use Case which are explained in the
        :ref:`user guide <add_project_to_a_use_case>`.

        Parameters
        ----------
        entity : Union[UseCaseReferenceEntity, Project, Dataset, Application]
            An existing entity to be linked to this Use Case.
            Cannot be used if entity_type and entity_id are passed.
        entity_type : UseCaseEntityType
            The entity type of the entity to link to this Use Case. Cannot be used if entity is passed.
        entity_id : str
            The ID of the entity to link to this Use Case. Cannot be used if entity is passed.

        Returns
        -------
        use_case_reference_entity : UseCaseReferenceEntity
            The newly created reference link between this Use Case and the entity.
        """
        e_type = self._resolve_api_entity(entity_type)
        e_id = entity_id
        if entity and (e_type or entity_id):
            raise InvalidUsageError(
                "Can only accept either an entity, or an entity type and entity id."
            )
        if not entity and (not e_type or not entity_id):
            raise InvalidUsageError("Missing entity, or an entity type and entity id.")
        if entity:
            e_type, e_id = get_reference_entity_info(entity)
        path = f"{self._path}{self.id}/{e_type}/{e_id}/"
        r_data = self._client.post(path)
        return UseCaseReferenceEntity.from_server_data(r_data.json())

    def remove(
        self,
        entity: Optional[Union[UseCaseReferenceEntity, T]] = None,
        entity_type: Optional[UseCaseEntityType] = None,
        entity_id: Optional[str] = None,
    ) -> None:
        """
        Remove an entity from a Use Case. Can only accept either an entity or
        an entity type and entity ID, but not both.

        Parameters
        ----------
        entity : Union[UseCaseReferenceEntity, Project, Dataset, Application]
            An existing entity instance to be removed from a Use Case.
            Cannot be used if entity_type and entity_id are passed.
        entity_type : UseCaseEntityType
            The entity type of the entity to link to this Use Case. Cannot be used if entity is passed.
        entity_id : str
            The ID of the entity to link to this Use Case.  Cannot be used if entity is passed.
        """
        e_type = self._resolve_api_entity(entity_type)
        e_id = entity_id
        if entity and (entity_type or entity_id):
            raise InvalidUsageError(
                "Can only accept either an entity, or an entity type and entity id."
            )
        if not entity and (not entity_type or not entity_id):
            raise InvalidUsageError("Missing entity, or an entity type and entity id.")
        if entity:
            e_type, e_id = get_reference_entity_info(entity)
        path = f"{self._path}{self.id}/{e_type}/{e_id}/"
        self._client.delete(path)
        return None

    def share(
        self,
        roles: List[SharingRole],
    ) -> None:
        """
        Share this Use Case with or remove access from one or more user(s).

        Parameters
        ----------
        roles : List[SharingRole]
            A list of :class:`SharingRole <datarobot.models.sharing.SharingRole>` instances, each of which
            references a user and a role to be assigned.

            Currently, the only supported roles for Use Cases are OWNER, EDITOR, and CONSUMER,
            and the only supported SHARING_RECIPIENT_TYPE is USER.

            To remove access, set a user's role to ``datarobot.enums.SHARING_ROLE.NO_ROLE``.

        Examples
        --------
        The :class:`SharingRole <datarobot.models.sharing.SharingRole>` class is needed in order to
        share a Use Case with one or more users.

        For example, suppose you had a list of user IDs you wanted to share this Use Case with. You could use
        a loop to generate a list of :class:`SharingRole <datarobot.models.sharing.SharingRole>` objects for them,
        and bulk share this Use Case.

        .. code-block:: python

            >>> from datarobot.models.use_cases.use_case import UseCase
            >>> from datarobot.models.sharing import SharingRole
            >>> from datarobot.enums import SHARING_ROLE, SHARING_RECIPIENT_TYPE
            >>>
            >>> user_ids = ["60912e09fd1f04e832a575c1", "639ce542862e9b1b1bfa8f1b", "63e185e7cd3a5f8e190c6393"]
            >>> sharing_roles = []
            >>> for user_id in user_ids:
            ...     new_sharing_role = SharingRole(
            ...         role=SHARING_ROLE.CONSUMER,
            ...         share_recipient_type=SHARING_RECIPIENT_TYPE.USER,
            ...         id=user_id,
            ...         can_share=True,
            ...     )
            ...     sharing_roles.append(new_sharing_role)
            >>> use_case = UseCase.get(use_case_id="5f33f1fd9071ae13568237b2")
            >>> use_case.share(roles=sharing_roles)

        Similarly, a :class:`SharingRole <datarobot.models.sharing.SharingRole>` instance can be used to
        remove a user's access if the ``role`` is set to ``SHARING_ROLE.NO_ROLE``, like in this example:

        .. code-block:: python

            >>> from datarobot.models.use_cases.use_case import UseCase
            >>> from datarobot.models.sharing import SharingRole
            >>> from datarobot.enums import SHARING_ROLE, SHARING_RECIPIENT_TYPE
            >>>
            >>> user_to_remove = "foo.bar@datarobot.com"
            ... remove_sharing_role = SharingRole(
            ...     role=SHARING_ROLE.NO_ROLE,
            ...     share_recipient_type=SHARING_RECIPIENT_TYPE.USER,
            ...     username=user_to_remove,
            ...     can_share=False,
            ... )
            >>> use_case = UseCase.get(use_case_id="5f33f1fd9071ae13568237b2")
            >>> use_case.share(roles=[remove_sharing_role])
        """
        if any(
            role.role
            not in [
                SHARING_ROLE.NO_ROLE,
                SHARING_ROLE.OWNER,
                SHARING_ROLE.EDITOR,
                SHARING_ROLE.CONSUMER,
            ]
            for role in roles
        ):
            raise InvalidUsageError(
                "Only NO_ROLE, OWNER, EDITOR, and CONSUMER roles are supported by Use Cases"
            )
        if any(role.share_recipient_type != SHARING_RECIPIENT_TYPE.USER for role in roles):
            raise InvalidUsageError(
                "Use Cases currently only support sharing with users, not organizations."
            )
        formatted_roles = [role.collect_payload() for role in roles]
        payload = {"roles": formatted_roles, "operation": "updateRoles"}
        path = f"{self._path}{self.id}/sharedRoles/"
        self._client.patch(path, data=payload)
        return None

    def get_shared_roles(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        id: Optional[str] = None,
    ) -> List[SharingRole]:
        """
        Retrieve access control information for this Use Case.

        Parameters
        ----------
        offset : Optional[int]
            The number of records to skip over. Optional. Default is 0.
        limit: Optional[int]
            The number of records to return. Optional. Default is 100.
        id: Optional[str]
            Return the access control information for a user with this user ID. Optional.
        """
        params = {"offset": offset, "limit": limit, "id": id}
        path = f"{self._path}{self.id}/sharedRoles/"
        r_data = unpaginate(path, params, self._client)
        return [SharingRole.from_server_data(data) for data in r_data]

    def list_projects(self) -> List[T]:
        """
        List all projects associated with this Use Case.

        Returns
        -------
        projects : List[Project]
            All projects associated with this Use Case.
        """
        from ..project import Project

        return Project.list(search_params={"experiment_container_ids": self.id})  # type: ignore[return-value]

    def list_datasets(self) -> List[T]:
        """
        List all datasets associated with this Use Case.

        Returns
        -------
        datasets : List[Dataset]
            All datasets associated with this Use Case.
        """
        from ..dataset import Dataset

        return Dataset.list(use_cases=self.id)  # type: ignore[return-value]

    def list_applications(self) -> List[T]:
        """
        List all applications associated with this Use Case.

        Returns
        -------
        applications : List[Application]
            All applications associated with this Use Case.
        """
        from ..application import Application

        return Application.list(use_cases=self.id)  # type: ignore[return-value]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import namedtuple
import itertools
from urllib.parse import parse_qs, urlparse

import pandas as pd
import trafaret as t

from datarobot import enums, errors
from datarobot._compat import Int, String

from ..utils import from_api
from .api_object import APIObject

_FEATURE = "Explanation_{}_feature_name"
_FEATURE_VAL = "Explanation_{}_feature_value"
_STRENGTH = "Explanation_{}_strength"


class RowsIterator:
    """
    Yields all available records one-by-one.

    While iterating, fetches rows from API with a series of requests with specified limit
    Stops iterating when API has no more objects to fetch
    """

    _trafaret = t.Dict(
        {t.Key("data"): t.List(t.Dict().allow_extra("*")), t.Key("next"): t.Or(t.String, t.Null)}
    ).allow_extra("*")

    def __init__(self, client, path, limit=None):
        self.client = client
        self.path = path
        self.query_params = dict(limit=limit)

        self._queue = []
        self._stop_flag = None

    def __iter__(self):
        return self

    def next(self):  # pylint: disable=missing-function-docstring
        if self._stop_flag and not self._queue:
            raise StopIteration()

        if not self._queue:
            items, next_url = self._load_more()
            self._queue.extend(items)
            if not self._queue:
                raise StopIteration()

            self._stop_flag = next_url is None
            self._stop_flag |= self.is_exhausted(len(items), self.query_params.get("limit"))
            if next_url is not None:
                self._save_next_url(next_url)

        data_dict = self._queue.pop(0)
        return data_dict

    __next__ = next

    @staticmethod
    def is_exhausted(loaded_count, limit):
        if limit is None:
            # If there was no limit param, assume all items were fetched
            return True

        # If we asked for 20 items but got 10 items that means our collection is exhausted
        return loaded_count < limit

    def _load_more(self):
        """Fetch more items from API, validate response, save next url"""
        response = self.client.get(self.path, params=self.query_params)
        if response.status_code != 200:
            e_msg = "Server returned unexpected status code"
            raise errors.ServerError(
                e_msg.format(response.status_code),
                response.status_code,
            )

        body = self._trafaret.check(response.json())
        return body["data"], body["next"]

    def _save_next_url(self, next_url):
        """Next URL should contain `limit` & `offset` query params"""
        parsed = urlparse(next_url)
        parsed_qs = parse_qs(parsed.query)
        self.query_params = {key: int(values_list[0]) for key, values_list in parsed_qs.items()}


TrainingPredictionsRow = namedtuple(
    "TrainingPredictionsRow",
    (
        "row_id,partition_id,prediction,prediction_values,"
        "timestamp,forecast_point,forecast_distance,series_id,"
        "prediction_explanations,shap_metadata"
    ),
)


class TrainingPredictionsIterator(RowsIterator):
    """
    Lazily fetches training predictions from DataRobot API in chunks of specified size and then
    iterates rows from responses as named tuples. Each row represents a training prediction
    computed for a dataset's row. Each named tuple has the following structure:

    Attributes
    ----------
    row_id : int
        id of the record in original dataset for which training prediction is calculated
    partition_id : str or float
        id of the data partition that the row belongs to. "0.0" corresponds to the validation
        partition or backtest 1.
    prediction : float
        the model's prediction for this data row
    prediction_values : list of dictionaries
        an array of dictionaries with a schema described as ``PredictionValue``
    timestamp : str or None
        (New in version v2.11) an ISO string representing the time of the prediction
        in time series project; may be None for non-time series projects
    forecast_point : str or None
        (New in version v2.11) an ISO string representing the point in time
        used as a basis to generate the predictions in time series project;
        may be None for non-time series projects
    forecast_distance : str or None
        (New in version v2.11) how many time steps are between the forecast point and the
        timestamp in time series project; None for non-time series projects
    series_id : str or None
        (New in version v2.11) the id of the series in a multiseries project;
        may be NaN for single series projects; None for non-time series projects
    prediction_explanations : list of dict or None
        (New in version v2.21) The prediction explanations for each feature. The total elements in
        the array are bounded by ``max_explanations`` and feature count. Only present if prediction
        explanations were requested. Schema described as ``PredictionExplanations``.
    shap_metadata : dict or None
        (New in version v2.21) The additional information necessary to understand SHAP based
        prediction explanations. Only present if `explanation_algorithm` equals
        `datarobot.enums.EXPLANATIONS_ALGORITHM.SHAP` was added in compute request. Schema
        described as ``ShapMetadata``.

    Notes
    -----
    Each ``PredictionValue`` dict contains these keys:

        label
            describes what this model output corresponds to. For regression
            projects, it is the name of the target feature. For classification and multiclass
            projects, it is a label from the target feature.
        value
            the output of the prediction. For regression projects, it is the
            predicted value of the target. For classification and multiclass projects, it is
            the predicted probability that the row belongs to the class identified by the label.

    Each ``PredictionExplanations`` dictionary contains these keys:

        label : string
            describes what output was driven by this prediction explanation. For regression
            projects, it is the name of the target feature. For classification projects, it is the
            class whose probability increasing would correspond to a positive strength of this
            prediction explanation.
        feature : string
            the name of the feature contributing to the prediction
        feature_value : object
            the value the feature took on for this row. The type corresponds to the feature
            (boolean, integer, number, string)
        strength : float
            algorithm-specific explanation value attributed to feature in this row

    ``ShapMetadata`` dictionary contains these keys:

        shap_remaining_total : float
            The total of SHAP values for features beyond the ``max_explanations``. This can be
            identically 0 in all rows, if `max_explanations` is greater than the number of features
            and thus all features are returned.
        shap_base_value : float
            the model's average prediction over the training data. SHAP values are deviations from
            the base value.
        warnings : dict or None
            SHAP values calculation warnings (e.g. additivity check failures in XGBoost models).
            Schema described as ``ShapWarnings``.

    ``ShapWarnings`` dictionary contains these keys:

        mismatch_row_count : int
            the count of rows for which additivity check failed
        max_normalized_mismatch : float
            the maximal relative normalized mismatch value

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # Fetch existing training predictions by their id
        training_predictions = dr.TrainingPredictions.get(project_id, prediction_id)

        # Iterate over predictions
        for row in training_predictions.iterate_rows()
            print(row.row_id, row.prediction)
    """

    _row_trafaret = t.Dict(
        {
            t.Key("row_id"): Int(),
            t.Key("partition_id"): t.Or(String(), t.Float()),
            t.Key("prediction"): t.Or(t.Float(), t.String()),
            t.Key("prediction_values"): t.List(
                t.Dict(
                    {t.Key("label"): t.Or(t.Float, t.String), t.Key("value"): t.Float}
                ).ignore_extra("*")
            ),
            t.Key("timestamp", default=None): t.Or(String, t.Null),
            t.Key("forecast_point", default=None): t.Or(String, t.Null),
            t.Key("forecast_distance", default=None): t.Or(Int, t.Null),
            t.Key("series_id", default=None): t.Or(String, Int, t.Null),
            t.Key("prediction_explanations", default=None): t.List(
                t.Dict(
                    {
                        t.Key("feature"): String(),
                        t.Key("feature_value"): String(),
                        t.Key("strength"): t.Float(),
                        t.Key("label"): t.Or(t.Float(), String()),
                    }
                )
            )
            | t.Null(),
            t.Key("shap_metadata", optional=True, default=None): t.Dict(
                {
                    t.Key("shap_base_value"): t.Float(),
                    t.Key("shap_remaining_total"): t.Float(),
                    t.Key("warnings", optional=True): t.Dict(
                        {
                            t.Key("mismatch_row_count"): Int(),
                            t.Key("max_normalized_mismatch"): t.Float(),
                        }
                    ),
                }
            )
            | t.Null(),
        }
    ).ignore_extra("*")

    def next(self):
        row_dict = super().next()
        row_dict = self._row_trafaret.check(row_dict)
        return TrainingPredictionsRow(**row_dict)

    __next__ = next

    def _load_more(self):
        rows, next_url = super()._load_more()
        return from_api(rows), next_url


class TrainingPredictions(APIObject):
    """
    Represents training predictions metadata and provides access to prediction results.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model
    prediction_id : str
        id of generated predictions
    data_subset : datarobot.enums.DATA_SUBSET
        data set definition used to build predictions.
        Choices are:

        - `datarobot.enums.DATA_SUBSET.ALL`
            for all data available. Not valid for models in datetime partitioned projects.
        - `datarobot.enums.DATA_SUBSET.VALIDATION_AND_HOLDOUT`
            for all data except training set. Not valid for models in datetime partitioned projects.
        - `datarobot.enums.DATA_SUBSET.HOLDOUT`
            for holdout data set only.
        - `datarobot.enums.DATA_SUBSET.ALL_BACKTESTS`
            for downloading the predictions for all backtest validation folds.
            Requires the model to have successfully scored all backtests.
            Datetime partitioned projects only.
    explanation_algorithm : datarobot.enums.EXPLANATIONS_ALGORITHM
        (New in version v2.21) Optional. If set to shap, the response will include prediction
        explanations based on the SHAP explainer (SHapley Additive exPlanations). Defaults to null
        (no prediction explanations).
    max_explanations : int
        (New in version v2.21) The number of top contributors that are included in prediction
        explanations. Max 100. Defaults to null for datasets narrower than 100 columns, defaults to
        100 for datasets wider than 100 columns.
    shap_warnings : list
        (New in version v2.21) Will be present if ``explanation_algorithm`` was set to
        `datarobot.enums.EXPLANATIONS_ALGORITHM.SHAP` and there were additivity failures during SHAP
        values calculation.

    Notes
    -----
    Each element in ``shap_warnings`` has the following schema:

    partition_name : str
        the partition used for the prediction record.
    value : object
        the warnings related to this partition.

    The objects in ``value`` are:

    mismatch_row_count : int
        the count of rows for which additivity check failed.
    max_normalized_mismatch : float
        the maximal relative normalized mismatch value.

    Examples
    --------
    Compute training predictions for a model on the whole dataset

    .. code-block:: python

        import datarobot as dr

        # Request calculation of training predictions
        training_predictions_job = model.request_training_predictions(dr.enums.DATA_SUBSET.ALL)
        training_predictions = training_predictions_job.get_result_when_complete()
        print('Training predictions {} are ready'.format(training_predictions.prediction_id))

        # Iterate over actual predictions
        for row in training_predictions.iterate_rows():
            print(row.row_id, row.partition_id, row.prediction)

    List all training predictions for a project

    .. code-block:: python

        import datarobot as dr

        # Fetch all training predictions for a project
        all_training_predictions = dr.TrainingPredictions.list(project_id)

        # Inspect all calculated training predictions
        for training_predictions in all_training_predictions:
            print(
                'Prediction {} is made for data subset "{}"'.format(
                    training_predictions.prediction_id,
                    training_predictions.data_subset,
                )
            )

    Retrieve training predictions by id

    .. code-block:: python

        import datarobot as dr

        # Getting training predictions by id
        training_predictions = dr.TrainingPredictions.get(project_id, prediction_id)

        # Iterate over actual predictions
        for row in training_predictions.iterate_rows():
            print(row.row_id, row.partition_id, row.prediction)
    """

    def __init__(
        self,
        project_id,
        prediction_id,
        model_id=None,
        data_subset=None,
        explanation_algorithm=None,
        max_explanations=None,
        shap_warnings=None,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.prediction_id = prediction_id
        self.path = self.build_path(project_id, prediction_id)
        self.data_subset = data_subset
        self.explanation_algorithm = explanation_algorithm
        self.max_explanations = max_explanations
        self.shap_warnings = shap_warnings

    @staticmethod
    def build_path(project_id, prediction_id=None):
        if prediction_id is not None:
            return f"projects/{project_id}/trainingPredictions/{prediction_id}/"

        return f"projects/{project_id}/trainingPredictions/"

    @classmethod
    def from_location(  # pylint: disable=arguments-renamed
        cls,
        location,
        data_subset=None,
        model_id=None,
        explanation_algorithm=None,
        max_explanations=None,
        shap_warnings=None,
    ):
        head, tail = location.split("/trainingPredictions/", 1)
        project_id, prediction_id = head.split("/")[-1], tail.split("/")[0]
        return cls(
            project_id,
            prediction_id,
            data_subset=data_subset,
            model_id=model_id,
            explanation_algorithm=explanation_algorithm,
            max_explanations=max_explanations,
            shap_warnings=shap_warnings,
        )

    @classmethod
    def list(cls, project_id):
        """
        Fetch all the computed training predictions for a project.

        Parameters
        ----------
        project_id : str
            id of the project

        Returns
        -------
        A list of :py:class:`TrainingPredictions` objects
        """
        _trafaret = t.Dict(
            {
                t.Key("data"): t.List(
                    t.Dict(
                        {
                            t.Key("url"): String(),
                            t.Key("model_id"): String(),
                            t.Key("data_subset"): String(),
                            t.Key("explanation_algorithm", optional=True): t.Or(String(), t.Null()),
                            t.Key("max_explanations", optional=True): t.Or(Int(), t.Null()),
                            t.Key("shap_warnings", optional=True): t.Dict(
                                {
                                    t.Key("partition_name"): String(),
                                    t.Key("value"): t.Dict(
                                        {
                                            t.Key("mismatch_row_count"): Int(),
                                            t.Key("max_normalized_mismatch"): t.Float(),
                                        }
                                    ),
                                }
                            )
                            | t.Null(),
                        }
                    ).ignore_extra("*")
                ),
            }
        ).ignore_extra("*")

        path = cls.build_path(project_id)
        converted = from_api(cls._server_data(path), keep_null_keys=True)
        validated = _trafaret.check(converted)["data"]
        return [
            cls.from_location(
                item["url"],
                data_subset=item["data_subset"],
                model_id=item["model_id"],
                explanation_algorithm=item.get("explanation_algorithm"),
                max_explanations=item.get("max_explanations"),
                shap_warnings=item.get("shap_warnings"),
            )
            for item in validated
        ]

    @classmethod
    def get(cls, project_id, prediction_id):
        """
        Retrieve training predictions on a specified data set.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        prediction_id : str
            id of the prediction set

        Returns
        -------
        :py:class:`TrainingPredictions` object which is ready to operate with specified predictions
        """
        return cls(project_id, prediction_id)

    def iterate_rows(self, batch_size=None):
        """
        Retrieve training prediction rows as an iterator.

        Parameters
        ----------
        batch_size : int, optional
            maximum number of training prediction rows to fetch per request

        Returns
        -------
        iterator : :py:class:`TrainingPredictionsIterator`
            an iterator which yields named tuples representing training prediction rows
        """
        return TrainingPredictionsIterator(self._client, self.path, limit=batch_size)

    def get_all_as_dataframe(self, class_prefix=enums.PREDICTION_PREFIX.DEFAULT, serializer="json"):
        """
        Retrieve all training prediction rows and return them as a pandas.DataFrame.

        Returned dataframe has the following structure:
            - row_id : row id from the original dataset
            - prediction : the model's prediction for this row
            - class_<label> : the probability that the target is this class (only appears for
              classification and multiclass projects)
            - timestamp : the time of the prediction (only appears for out of time validation or
              time series projects)
            - forecast_point : the point in time used as a basis to generate the predictions
              (only appears for time series projects)
            - forecast_distance : how many time steps are between timestamp and forecast_point
              (only appears for time series projects)
            - series_id : he id of the series in a multiseries project
              or None for a single series project
              (only appears for time series projects)

        Parameters
        ----------
        class_prefix : str, optional
            The prefix to append to labels in the final dataframe. Default is ``class_``
            (e.g., apple -> class_apple)
        serializer : str, optional
            Serializer to use for the download. Options: ``json`` (default) or ``csv``.

        Returns
        -------
        dataframe: pandas.DataFrame
        """

        serializers = {
            "json": self._get_all_as_dataframe_json,
            "csv": self._get_all_as_dataframe_csv,
        }

        if serializer not in serializers:
            raise ValueError(f'Unknown serializer "{serializer}", use "json" or "csv"')

        return serializers[serializer](class_prefix)

    def _get_all_as_dataframe_json(
        self, class_prefix
    ):  # pylint: disable=missing-function-docstring
        rows = self.iterate_rows()

        tmp, rows = itertools.tee(rows)
        first_row = next(tmp, None)
        is_classification = first_row is not None and len(first_row.prediction_values) > 1
        is_datetime_partitioned = self._is_datetime_partitioned

        if is_classification:
            labels = (p["label"] for p in first_row.prediction_values)
            columns = self._get_classification_columns(
                labels,
                class_prefix=class_prefix,
                is_datetime_partitioned=is_datetime_partitioned,
                prediction_explanations=first_row.prediction_explanations,
            )
            return self._build_classification_dataframe(rows, columns, is_datetime_partitioned)
        elif self._is_time_series_project:
            return self._build_timeseries_dataframe(rows)
        else:
            return self._build_regression_dataframe(
                rows, is_datetime_partitioned, first_row.prediction_explanations
            )

    @property
    def _is_time_series_project(self):
        from datarobot.models import (  # pylint: disable=import-outside-toplevel,cyclic-import
            Project,
        )

        project = Project.get(self.project_id)
        return project.use_time_series

    @property
    def _is_datetime_partitioned(self):
        from datarobot.models import (  # pylint: disable=import-outside-toplevel,cyclic-import
            Project,
        )

        project = Project.get(self.project_id)
        return project.is_datetime_partitioned

    @staticmethod
    def _build_timeseries_dataframe(rows):  # pylint: disable=missing-function-docstring
        columns = (
            "row_id",
            "partition_id",
            "prediction",
            "timestamp",
            "forecast_point",
            "forecast_distance",
            "series_id",
        )
        data = (
            (
                row.row_id,
                row.partition_id,
                row.prediction,
                row.timestamp,
                row.forecast_point,
                row.forecast_distance,
                row.series_id,
            )
            for row in rows
        )
        return pd.DataFrame.from_records(data, columns=columns)

    @staticmethod
    def _build_regression_dataframe(
        rows, is_datetime_partitioned, prediction_explanations
    ):  # pylint: disable=missing-function-docstring
        if is_datetime_partitioned:
            columns = (
                "row_id",
                "partition_id",
                "prediction",
                "timestamp",
            )
        else:
            columns = (
                "row_id",
                "partition_id",
                "prediction",
            )
        if prediction_explanations:
            for i in range(len(prediction_explanations)):
                idx = i + 1
                columns += (
                    _FEATURE.format(idx),
                    _FEATURE_VAL.format(idx),
                    _STRENGTH.format(idx),
                )
            columns += (
                "shap_remaining_total",
                "shap_base_value",
            )

        data = []
        for row in rows:
            data_row = (
                row.row_id,
                row.partition_id,
                row.prediction,
            )
            if is_datetime_partitioned:
                data_row += (row.timestamp,)
            if row.prediction_explanations:
                for prediction_explanation in row.prediction_explanations:
                    data_row += (
                        prediction_explanation["feature"],
                        prediction_explanation["feature_value"],
                        prediction_explanation["strength"],
                    )
                data_row += (
                    row.shap_metadata["shap_remaining_total"],
                    row.shap_metadata["shap_base_value"],
                )
            data.append(data_row)
        return pd.DataFrame.from_records(data, columns=columns)

    @staticmethod
    def _build_classification_dataframe(
        rows, columns, is_datetime_partitioned
    ):  # pylint: disable=missing-function-docstring
        data_list = []
        for row in rows:
            data_row = (row.row_id, row.partition_id, row.prediction)
            data_row += tuple(prediction["value"] for prediction in row.prediction_values)
            if is_datetime_partitioned:
                data_row += (row.timestamp,)
            if row.prediction_explanations:
                data_row += (
                    row.prediction_explanations[0]["label"],
                    row.shap_metadata["shap_remaining_total"],
                    row.shap_metadata["shap_base_value"],
                )
                for prediction_explanation in row.prediction_explanations:
                    data_row += (
                        prediction_explanation["feature"],
                        prediction_explanation["feature_value"],
                        prediction_explanation["strength"],
                    )
            data_list.append(data_row)

        return pd.DataFrame(data_list, columns=columns)

    @staticmethod
    def _get_classification_columns(  # pylint: disable=missing-function-docstring
        class_labels, class_prefix, is_datetime_partitioned, prediction_explanations
    ):
        columns = ("row_id", "partition_id", "prediction")
        columns += tuple(f"{class_prefix}{label}" for label in class_labels)
        if is_datetime_partitioned:
            columns += ("timestamp",)
        if prediction_explanations:
            columns += (
                "explained_class",
                "shap_remaining_total",
                "shap_base_value",
            )
            for i in range(len(prediction_explanations)):
                idx = i + 1
                columns += (
                    _FEATURE.format(idx),
                    _FEATURE_VAL.format(idx),
                    _STRENGTH.format(idx),
                )
        return columns

    def _get_all_as_dataframe_csv(self, class_prefix):  # pylint: disable=unused-argument
        resp = self._client.get(self.path, headers={"Accept": "text/csv"}, stream=True)
        if resp.status_code == 200:
            return pd.read_csv(resp.raw)

        raise errors.ServerError(
            f"Server returned unknown status code: {resp.status_code}",
            resp.status_code,
        )

    def download_to_csv(self, filename, encoding="utf-8", serializer="json"):
        """
        Save training prediction rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save training prediction rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to
            'utf-8'
        serializer : str, optional
            Serializer to use for the download. Options: ``json`` (default) or ``csv``.
        """
        df = self.get_all_as_dataframe(serializer=serializer)
        df.to_csv(
            path_or_buf=filename,
            header=True,
            index=False,
            encoding=encoding,
        )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# pylint: disable=cyclic-import
import contextvars
import functools
from inspect import Parameter, signature
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, TypeVar, Union
import warnings

from typing_extensions import ParamSpec

from datarobot.enums import UseCaseReferenceEntityMap
from datarobot.errors import InvalidUsageError, MultipleUseCasesNotAllowed
from datarobot.models.use_cases.use_case import (
    get_reference_entity_info,
    UseCase,
    UseCaseReferenceEntity,
)

from ...context import Context
from ...utils.logger import get_logger

P = ParamSpec("P")
T = TypeVar("T")

UseCaseLike = Union[List[UseCase], UseCase, List[str], str]

_use_case_linked: contextvars.ContextVar[Optional[Set[Tuple[str, str]]]] = contextvars.ContextVar(
    "use_case_linked", default=None
)
_decorator_use_cases: contextvars.ContextVar[Optional[UseCaseLike]] = contextvars.ContextVar(
    "decorator_use_cases", default=None
)


logger = get_logger(__name__)


def resolve_use_cases(
    params: Dict[str, Any],
    use_cases: Optional[UseCaseLike] = None,
    use_case_key: str = "experiment_container_ids",
) -> Dict[str, Any]:
    """
    Add a global Use Case ID or IDs to query params for a list operation if no use_case_id has been passed
    This method supports checking for `use_case_id` in a params dict if users can build their own params dict.
    Parameters
    ----------
    params : Dict[str, Any]
        The query params dict to add a Use Case ID to, if a global ID exists, or if one was passed in directly.
    use_cases : Optional[Union[List[UseCase], UseCase, List[str], str]]
        Optional. The use case or cases to add to a query params dict. May be Use Case ID or the Use Case entity.
    use_case_key : Optional[str]
        Optional. The key that will be used in the query params for Use Case ID. Default is 'experiment_container_ids'.
    Returns
    -------
    params : Dict[str, Any]
        If a Use Case ID is available, the params with the ID added. Otherwise, return the dict unmodified.
    """
    # Check to see if a use_case_id is already in the params dict
    if not params.get(use_case_key):
        # If use_case_key is not in the dict, add in the manually passed value, if it exists
        if use_cases:
            params[use_case_key] = resolve_use_case_ids(use_cases)
        elif Context.use_case:
            params[use_case_key] = Context.use_case.id  # type: ignore[union-attr]
    return params


def resolve_use_case_ids(use_cases: UseCaseLike) -> List[str]:
    """
    A helper function to convert a list of UseCase objects, single UseCase, or single Use Case ID
    into a list of strings.

    Parameters
    ----------
    use_cases : List[UseCase], UseCase, List[str], str
        The list of UseCase objects, list of strings, single UseCase object, or single Use Case ID to turn into
        a list of ID strings.

    Returns
    -------
    use_case_ids : List[str]
        The returned list of ID strings.
    """
    if isinstance(use_cases, list):
        return [
            use_case.id if isinstance(use_case, UseCase) else use_case for use_case in use_cases
        ]
    if isinstance(use_cases, UseCase):
        return [use_cases.id]
    else:
        return [use_cases]


def _add_to_use_case(use_case_id: str, entity: Union[UseCaseReferenceEntity, T]) -> None:
    _, entity_id = get_reference_entity_info(entity)
    linked = _use_case_linked.get()
    if entity_id is not None and linked is not None and (use_case_id, entity_id) not in linked:
        use_case = UseCase.get(use_case_id=use_case_id)
        use_case.add(entity=entity)
        linked.add((use_case_id, entity_id))


def add_to_use_case(
    allow_multiple: bool,
) -> Callable[[Callable[P, T]], Callable[..., T]]:
    """
    A decorator to mark functions as adding the return value to a given Use Case. When implemented,
    the decorator will:

    1. add a `use_cases` keyword-only argument to the decorated function or method,
    2. add logic to automatically add the returned object to a Use Case, AFTER the decorated function
       has finished executing.

    This invokes the UseCase API and does NOT require the underlying route / SmartController to
    consume a use_case_id as an argument.

    Parameters
    ----------
    allow_multiple : bool
        Whether the function should be decorated to permit adding to multiple use cases. Default is False.

    Examples
    --------
    This decorator should only be added to methods that return a type listed in
    enums.UseCaseEntityType or enums.UseCaseExtendedEntityType.
    The function needs to have the return annotation added.

    This function could be decorated:
    .. code-block:: python
        def foo_bar() -> Dataset:
            ....
            return Dataset()

    These functions could not be decorated:
    .. code-block:: python
        def foo_bar():
            ....
            return Dataset()

        def foo_bar() -> Model:
            ....
            return Model()

    To decorate, it's as simple as adding the decorator and updating the doc string for the method
    or function with one of the following:

    use_case: UseCase | string, optional
            A single UseCase object or ID to add this new <return_object_type> to. Must be a kwarg.

    use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new <return_object_type> to. Must be a kwarg.


    So this:
    .. code-block:: python
        def foo_bar() -> Dataset:
            ....
            return Dataset()

    Would become this:
    .. code-block:: python
        @add_to_use_case(allow_multiple=True)
        def foo_bar() -> Dataset:
            \"\"\"
            use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new Dataset to. Must be a kwarg.
            \"\"\"
            ....
            return Dataset()

    Returns
    -------
    func : callable
        The wrapped function, with an additional kwarg 'use_case' or 'use_cases' depending on the
        value of the 'allow_multiple' param.
    """

    def wrapper(func: Callable[P, T]) -> Callable[..., T]:
        ret_sig = signature(func)
        ret_type = ret_sig.return_annotation
        # Check if return annotation is a string or a type. If it's a type, get the name and make it lowercase.
        # If the return annotation is a string, just make it lower case.
        # Some return types are "TDataset", others are Dataset<datarobot.Dataset>
        ret_type = ret_type.__name__.lower() if not isinstance(ret_type, str) else ret_type.lower()
        ret_type = ret_type.lstrip("t")
        if not UseCaseReferenceEntityMap.get(ret_type):
            raise InvalidUsageError(
                "This decorator can only support methods that return a "
                "Project, Dataset, or Application instance."
            )
        new_kw_param = Parameter(
            "use_cases" if allow_multiple else "use_case", kind=Parameter.KEYWORD_ONLY, default=None
        )
        param_list = list(ret_sig.parameters.values())
        param_list.append(new_kw_param)
        new_sig = ret_sig.replace(parameters=param_list)
        func.__signature__ = new_sig  # type: ignore[attr-defined]

        if allow_multiple:

            @functools.wraps(func)
            def add_to_multiple(
                *args: P.args,
                use_cases: Optional[UseCaseLike] = None,
                **kwargs: P.kwargs,
            ) -> T:
                """
                Parameters
                ----------
                use_cases: list[UseCase] | UseCase | list[string] | string, optional
                    A list of UseCase objects, UseCase object,
                    list of Use Case IDs or a single Use Case ID to add this new entity to. Must be a kwarg.
                """
                use_case_ids = None
                context_use_case_token = None
                linked_use_cases_token = None

                # take care of use_case param
                if use_cases is None:
                    # check contextvar that could be set by outer iterations first then Context
                    use_cases = _decorator_use_cases.get(Context.use_case)
                else:
                    # if use_case passed, set it to be used by inner calls instead of Context
                    context_use_case_token = _decorator_use_cases.set(use_cases)

                try:
                    if use_cases is not None:
                        # Should only ever resolve to a single use case
                        use_case_ids = resolve_use_case_ids(use_cases=use_cases)

                    # _use_case_linked never created? must be top level call, initialize set
                    # to track entities added to UseCase before we call the wrapped function
                    # as it can also call decorated method inside
                    if _use_case_linked.get() is None:
                        linked_use_cases_token = _use_case_linked.set(set())
                    ret = func(*args, **kwargs)

                    if use_case_ids:
                        for use_case_id in use_case_ids:
                            _add_to_use_case(use_case_id, ret)
                finally:
                    if context_use_case_token:
                        _decorator_use_cases.reset(context_use_case_token)
                    if linked_use_cases_token:
                        _use_case_linked.reset(linked_use_cases_token)

                return ret

            return add_to_multiple
        else:

            @functools.wraps(func)
            def add_to_one(
                *args: P.args,
                use_case: Optional[Union[UseCase, str]] = None,
                **kwargs: P.kwargs,
            ) -> T:
                """
                Parameters
                ----------
                use_case: UseCase | string, optional
                    A single UseCase object or ID to add this new <return_object_type> to. Must be a kwarg.
                """
                use_case_id: Optional[str] = None
                context_use_case_token: Optional[contextvars.Token[Union[UseCaseLike, None]]] = None
                linked_use_cases_token = None

                # take care of use_case param
                if use_case is None:
                    # check contextvar that could be set by outer iterations first then Context
                    use_cases = _decorator_use_cases.get(Context.use_case)
                    # but it may have been set to a list
                    if isinstance(use_cases, list):
                        warnings.warn(MultipleUseCasesNotAllowed())
                    else:
                        use_case = use_cases
                else:
                    # if use_case passed, set it to be used by inner calls instead of Context
                    context_use_case_token = _decorator_use_cases.set(use_case)

                try:
                    if use_case is not None:
                        # Should only ever resolve to a single use case
                        use_case_id = resolve_use_case_ids(use_cases=use_case)[0]

                    # _use_case_linked never created? must be top level call, initialize set
                    # to track entities added to UseCase before we call the wrapped function
                    # as it can also call decorated method inside
                    if _use_case_linked.get() is None:
                        linked_use_cases_token = _use_case_linked.set(set())
                    ret = func(*args, **kwargs)

                    if use_case_id:
                        _add_to_use_case(use_case_id, ret)
                finally:
                    if context_use_case_token:
                        _decorator_use_cases.reset(context_use_case_token)
                    if linked_use_cases_token:
                        _use_case_linked.reset(linked_use_cases_token)

                return ret

            return add_to_one

    return wrapper



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from functools import partial
import json
from typing import cast, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot._compat import String
from datarobot.enums import ComplianceDocTemplateType, ComplianceDocType
from datarobot.models.api_object import APIObject

# Make sure validator stays insync with main repo
OptKey = partial(t.Key, optional=True)
TitleString = String(max_length=500)
SectionDict = t.Forward()
SectionDict.provide(
    t.Or(
        t.Dict(
            {
                "type": t.Atom("datarobot"),
                "title": TitleString,
                "content_id": String(),
                OptKey("sections"): t.Or(t.List(SectionDict), t.Null),
                OptKey("description"): t.Or(String(allow_blank=True), t.Null),
                OptKey("instructions"): t.Or(
                    t.Dict({"owner": String(allow_blank=True), "user": String(allow_blank=True)}),
                    t.Null,
                ),
                OptKey("locked"): t.Bool,
            }
        ),
        t.Dict(
            {
                "type": t.Atom("user"),
                "title": TitleString,
                "regular_text": String(max_length=5000, allow_blank=True),
                "highlighted_text": String(max_length=5000, allow_blank=True),
                OptKey("sections"): t.Or(t.List(SectionDict), t.Null),
                OptKey("description"): t.Or(String(allow_blank=True), t.Null),
                OptKey("instructions"): t.Or(
                    t.Dict({"owner": String(allow_blank=True), "user": String(allow_blank=True)}),
                    t.Null,
                ),
                OptKey("locked"): t.Bool,
            }
        ),
        t.Dict(
            {
                "type": t.Atom("custom"),
                "title": TitleString,
                "regular_text": String(max_length=5000, allow_blank=True),
                "highlighted_text": String(max_length=5000, allow_blank=True),
                OptKey("sections"): t.Or(t.List(SectionDict), t.Null),
                OptKey("description"): t.Or(String(allow_blank=True), t.Null),
                OptKey("instructions"): t.Or(
                    t.Dict({"owner": String(allow_blank=True), "user": String(allow_blank=True)}),
                    t.Null,
                ),
                OptKey("locked"): t.Bool,
            }
        ),
        t.Dict(
            {"type": t.Atom("table_of_contents"), "title": TitleString, OptKey("locked"): t.Bool}
        ),
    )
)


class Instructions(TypedDict):
    owner: str
    user: str


class Section(TypedDict):
    title: str
    type: ComplianceDocType
    locked: bool


class DataRobotSection(Section):
    content_id: str
    sections: Optional[List["Section"]]
    description: Optional[str]
    instructions: Optional[Instructions]


class UserSection(Section):
    regular_test: str
    highlighted_test: str
    sections: Optional[List["Section"]]
    description: Optional[str]
    instructions: Optional[Instructions]


class CustomSection(Section):
    regular_test: str
    highlighted_test: str
    sections: Optional[List["Section"]]
    description: Optional[str]
    instructions: Optional[Instructions]


class TableOfContentsSection(Section):
    """
    Table of Contents Sections only have the base fields available to all Section classes
    """


def make_url(root: str, template_id: str) -> str:
    return f"{root}{template_id}/"


def load_sections_from_path(path: str) -> List[Section]:
    with open(path) as f:  # pylint: disable=unspecified-encoding
        sections = json.loads(f.read())
        return cast(List[Section], sections)


class ComplianceDocTemplate(APIObject):
    """A :ref:`compliance documentation template <automated_documentation_overview>`. Templates
    are used to customize contents of :class:`AutomatedDocument
    <datarobot.models.automated_documentation.AutomatedDocument>`.

    .. versionadded:: v2.14

    Notes
    -----
    Each ``section`` dictionary has the following schema:

    * ``title`` : title of the section
    * ``type`` : type of section. Must be one of "datarobot", "user" or "table_of_contents".

    Each type of section has a different set of attributes described bellow.

    Section of type ``"datarobot"`` represent a section owned by DataRobot. DataRobot
    sections have the following additional attributes:

    * ``content_id`` : The identifier of the content in this section.
      You can get the default template with :meth:`get_default
      <datarobot.models.compliance_doc_template.ComplianceDocTemplate.get_default>`
      for a complete list of possible DataRobot section content ids.
    * ``sections`` :  list of sub-section dicts nested under the parent section.

    Section of type ``"user"`` represent a section with user-defined content.
    Those sections may contain text generated by user and have the following additional fields:

    * ``regularText`` : regular text of the section, optionally separated by
      ``\\n`` to split paragraphs.
    * ``highlightedText`` : highlighted text of the section, optionally separated
      by ``\\n`` to split paragraphs.
    * ``sections`` :  list of sub-section dicts nested under the parent section.

    Section of type ``"table_of_contents"`` represent a table of contents and has
    no additional attributes.

    Attributes
    ----------
    id : str
        the id of the template
    name : str
        the name of the template.
    creator_id : str
        the id of the user who created the template
    creator_username : str
        username of the user who created the template
    org_id : str
        the id of the organization the template belongs to
    sections : list of dicts
        the sections of the template describing the structure of the document. Section schema
        is described in Notes section above.
    """

    _root_path = "complianceDocTemplates/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("creator_id"): String(),
            t.Key("creator_username"): String(),
            OptKey("org_id"): t.Or(String(), t.Null),
            t.Key("name"): String(),
            OptKey("sections"): t.Or(t.List(SectionDict), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        creator_id: str,
        creator_username: str,
        name: str,
        org_id: Optional[str] = None,
        sections: Optional[List[Section]] = None,
    ):
        self.id = id
        self.creator_id = creator_id
        self.creator_username = creator_username
        self.org_id = org_id
        self.name = name
        self.sections = sections

    def __repr__(self) -> str:
        return f"ComplianceDocTemplate({self.name!r})"

    @classmethod
    def get_default(
        cls, template_type: Optional[ComplianceDocTemplateType] = None
    ) -> "ComplianceDocTemplate":
        """Get a default DataRobot template. This template is used for generating
        compliance documentation when no template is specified.


        Parameters
        ----------
        template_type : str or None
            Type of the template. Currently supported values are "normal" and "time_series"

        Returns
        -------
        template : ComplianceDocTemplate
            the default template object with ``sections`` attribute populated with default sections.
        """
        query_params = {"type": template_type} if template_type else None
        response = cls._client.get("complianceDocTemplates/default/", params=query_params)
        return cls(
            id=cast(str, None),
            creator_id=cast(str, None),
            creator_username=cast(str, None),
            org_id=None,
            name="default",
            sections=response.json()["sections"],
        )

    @classmethod
    def create_from_json_file(cls, name: str, path: str) -> "ComplianceDocTemplate":
        """Create a template with the specified name and sections in a JSON file.

        This is useful when working with sections in a JSON file. Example:

        .. code-block:: python

            default_template = ComplianceDocTemplate.get_default()
            default_template.sections_to_json_file('path/to/example.json')
            # ... edit example.json in your editor
            my_template = ComplianceDocTemplate.create_from_json_file(
                name='my template',
                path='path/to/example.json'
            )

        Parameters
        ----------
        name : str
            the name of the template. Must be unique for your user.
        path : str
            the path to find the JSON file at

        Returns
        -------
        template : ComplianceDocTemplate
            the created template
        """
        resp = cls._client.post(
            cls._root_path, data={"name": name, "sections": load_sections_from_path(path)}
        )
        return cls.from_location(resp.headers["Location"])

    @classmethod
    def create(cls, name: str, sections: List[Section]) -> "ComplianceDocTemplate":
        """Create a template with the specified name and sections.

        Parameters
        ----------
        name : str
            the name of the template. Must be unique for your user.
        sections : list
            list of section objects

        Returns
        -------
        template : ComplianceDocTemplate
            the created template
        """
        resp = cls._client.post(cls._root_path, data={"name": name, "sections": sections})
        return cls.from_location(resp.headers["Location"])

    @classmethod
    def get(cls, template_id: str) -> "ComplianceDocTemplate":
        """Retrieve a specific template.

        Parameters
        ----------
        template_id :  str
            the id of the template to retrieve

        Returns
        -------
        template : ComplianceDocTemplate
            the retrieved template
        """
        return cls.from_location(make_url(cls._root_path, template_id))

    @classmethod
    def list(
        cls,
        name_part: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List["ComplianceDocTemplate"]:
        """Get a paginated list of compliance documentation template objects.

        Parameters
        ----------
        name_part : str or None
            Return only the templates with names matching specified string. The matching is
            case-insensitive.
        limit : int
            The number of records to return. The server will use a (possibly finite) default if not
            specified.
        offset : int
            The number of records to skip.

        Returns
        -------
        templates : list of ComplianceDocTemplate
            the list of template objects
        """
        r_data = cls._client.get(
            cls._root_path, params={"limit": limit, "offset": offset, "namePart": name_part}
        ).json()
        return cast(
            List["ComplianceDocTemplate"], [cls.from_server_data(item) for item in r_data["data"]]
        )

    def sections_to_json_file(self, path: str, indent: int = 2) -> None:
        """Save sections of the template to a json file at the specified path

        Parameters
        ----------
        path : str
            the path to save the file to
        indent : int
            indentation to use in the json file.
        """
        with open(path, "w") as f:  # pylint: disable=unspecified-encoding
            f.write(json.dumps(self.sections, indent=indent))

    def update(self, name: Optional[str] = None, sections: Optional[List[Section]] = None) -> None:
        """Update the name or sections of an existing doc template.

        Note that default or non-existent templates can not be updated.

        Parameters
        ----------
        name : str, optional
            the new name for the template
        sections : list of dicts
            list of sections
        """
        payload: Dict[str, Union[str, List[Section]]] = {}
        if name is not None:
            payload["name"] = name
        if sections is not None:
            payload["sections"] = sections

        self._client.patch(make_url(self._root_path, self.id), data=payload)

        if name is not None:
            self.name = name
        if sections is not None:
            self.sections = sections

    def delete(self) -> None:
        """Delete the compliance documentation template."""
        self._client.delete(make_url(self._root_path, self.id))
